from pydiverse.pipedag.core import Flow, Stage, Task
from pydiverse.pipedag.materialize import Blob, Table, materialize

__all__ = [
    "Blob",
    "Flow",
    "Stage",
    "Table",
    "materialize",
]



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import structlog
from attrs import frozen

from pydiverse.pipedag.context import ConfigContext, StageLockContext
from pydiverse.pipedag.context.run_context import DematerializeRunContext
from pydiverse.pipedag.core.task import Task, TaskGetItem
from pydiverse.pipedag.errors import LockError
from pydiverse.pipedag.materialize.core import MaterializingTask
from pydiverse.pipedag.util import deep_map

if TYPE_CHECKING:
    from pydiverse.pipedag._typing import Materializable


@frozen
class Result:
    underlying: Any
    successful: bool
    config_context: ConfigContext | None

    task_values: dict[Task, Any]
    exception: Exception = None

    def get(self, task: Task | TaskGetItem, as_type: type = None) -> Materializable:
        """Load the results of a task from the database.

        :param task: The task
        :param as_type: The type as which tables produced by this task should
            be dematerialized. If no type is specified, the input type of
            the task is used.
        :return: The results of the task.
        """
        if not self.successful:
            logger = structlog.getLogger()
            logger.warning(
                "Attention: getting tables from unsuccessful run is unreliable!"
            )

        if isinstance(task, Task):
            root_task = task
        else:
            root_task = task.task

        if as_type is None:
            assert isinstance(root_task, MaterializingTask)
            as_type = root_task.input_type

        if self.config_context.strict_result_get_locking:
            try:
                StageLockContext.get()
            except LookupError:
                raise LockError(
                    "Called Result.get() without opening StageLockContext. Consider"
                    " using 'strict_result_get_locking: false' for interactive"
                    " debugging"
                ) from None

        # TODO: Check that the results loaded from the database correspond
        #       to the run_id of this result object.
        with self.config_context as config_ctx, DematerializeRunContext() as run_ctx:

            def dematerialize_mapper(item):
                return config_ctx.store.dematerialize_item(
                    item, as_type=as_type, ctx=run_ctx
                )

            task_value = self.task_values[root_task]
            return deep_map(task.resolve_value(task_value), dematerialize_mapper)



"""Utilities for JSON to support PipeDAG objects

PipeDAG objects get serialized as JSON objects with a special `_pipedag_type_`
key to identify them when decoding. The associated value encodes the
type of the object.
"""

from __future__ import annotations

from pathlib import Path

from pydiverse.pipedag import Stage
from pydiverse.pipedag.context import RunContext
from pydiverse.pipedag.materialize.container import Blob, RawSql, Table
from pydiverse.pipedag.util.config import PipedagConfig

PIPEDAG_TYPE = "_pipedag_type_"
PIPEDAG_TYPE_TABLE = "table"
PIPEDAG_TYPE_RAWSQL = "raw_sql"
PIPEDAG_TYPE_BLOB = "blob"
PIPEDAG_TYPE_STAGE = "stage"
PIPEDAG_TYPE_PIPEDAG_CONFIG = "pipedag_config"
PIPEDAG_TYPE_PATH = "path"


def json_default(o):
    """Encode `Table`, `RawSql`, 'Stage', and `Blob` objects"""
    if isinstance(o, Table):
        return {
            PIPEDAG_TYPE: PIPEDAG_TYPE_TABLE,
            "stage": o.stage.name,
            "name": o.name,
            "primary_key": o.primary_key,
            "indexes": o.indexes,
            "cache_key": o.cache_key,
        }
    if isinstance(o, RawSql):
        return {
            PIPEDAG_TYPE: PIPEDAG_TYPE_RAWSQL,
            "stage": o.stage.name,
            "name": o.name,
            "cache_key": o.cache_key,
        }
    if isinstance(o, Blob):
        return {
            PIPEDAG_TYPE: PIPEDAG_TYPE_BLOB,
            "stage": o.stage.name,
            "name": o.name,
            "cache_key": o.cache_key,
        }
    if isinstance(o, Stage):
        return {
            PIPEDAG_TYPE: PIPEDAG_TYPE_STAGE,
            "name": o.name,
        }
    if isinstance(o, PipedagConfig):
        return {
            PIPEDAG_TYPE: PIPEDAG_TYPE_PIPEDAG_CONFIG,
            "config_dict": o.config_dict,
        }
    if isinstance(o, Path):
        return {
            PIPEDAG_TYPE: PIPEDAG_TYPE_PATH,
            "path": str(o),
        }

    raise TypeError(f"Object of type {type(o).__name__} is not JSON serializable")


def json_object_hook(d: dict):
    """Decode json with `Table` and `Blob` objects"""
    pipedag_type = d.get(PIPEDAG_TYPE)
    if pipedag_type:
        run_context = RunContext.get()
        stages = run_context.flow.stages

        if pipedag_type == PIPEDAG_TYPE_TABLE:
            tbl = Table(
                name=d["name"],
                stage=stages[d["stage"]],
                primary_key=d["primary_key"],
                indexes=d["indexes"],
            )
            tbl.cache_key = d["cache_key"]
            return tbl
        elif pipedag_type == PIPEDAG_TYPE_RAWSQL:
            raw_sql = RawSql(
                name=d["name"],
                stage=stages[d["stage"]],
            )
            raw_sql.cache_key = d["cache_key"]
            return raw_sql
        elif pipedag_type == PIPEDAG_TYPE_BLOB:
            blob = Blob(
                name=d["name"],
                stage=stages[d["stage"]],
            )
            blob.cache_key = d["cache_key"]
            return blob
        elif pipedag_type == PIPEDAG_TYPE_STAGE:
            return stages[d["name"]]
        elif pipedag_type == PIPEDAG_TYPE_PIPEDAG_CONFIG:
            # PipedagConfig objects are allowed as input to @materialize tasks,
            # but it is not allowed as output since this might cause trouble
            # for cache-invalidation
            raise TypeError("PipedagConfig can't be deserialized.")
        elif pipedag_type == PIPEDAG_TYPE_PATH:
            return Path(d["path"])
        else:
            raise ValueError(
                f"Invalid value for '{PIPEDAG_TYPE}' key: {repr(pipedag_type)}"
            )

    return d



from __future__ import annotations

from typing import TYPE_CHECKING, Generic

from pydiverse.pipedag._typing import T
from pydiverse.pipedag.util import normalize_name

if TYPE_CHECKING:
    from pydiverse.pipedag.core.stage import Stage


class Table(Generic[T]):
    """Table container

    Used to wrap table objects that get returned from materializing
    tasks.

    :param obj: The table object to wrap
    :param name: Optional name. If no name is provided, an automatically
        generated name will be used. To prevent name collisions, you can
        add '%%' at the end of the name to enable automatic name mangling.
    :param primary_key: Optional name of the primary key that should be
        used when materializing this table
    """

    def __init__(
        self,
        obj: T | None = None,
        name: str | None = None,
        stage: Stage | None = None,
        primary_key: str | list[str] | None = None,
        indexes: list[list[str]] | None = None,
    ):
        self._name = None

        self.obj = obj
        self.name = name
        self.stage = stage
        self.primary_key = primary_key
        self.indexes = indexes

        # cache_key will be overridden shortly before handing over to downstream tasks
        # that use it to compute their input_hash for cache_invalidation due to input
        # change
        self.cache_key = None

    def __str__(self):
        return f"<Table: {self.name} ({self.stage.name})>"

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if value is not None and not isinstance(value, str):
            raise TypeError(f"Table name must be of instance 'str' not {type(value)}.")
        self._name = normalize_name(value)


class RawSql:
    """Container for raw sql strings

    This allows wrapping legacy sql code with pipedag before it is converted
    to proper tasks that allow tracing tables.

    :param sql: The table object to wrap
    :param cache_key: Internal cache_key used when retreiving an object
        from the database cache.
    """

    def __init__(
        self,
        sql: str | None = None,
        name: str | None = None,
        stage: Stage | None = None,
    ):
        self._name = None

        self.sql = sql
        self.name = name
        self.stage = stage

        # cache_key will be overridden shortly before handing over to downstream tasks
        # that use it to compute their input_hash for cache_invalidation due to input
        # change
        self.cache_key = None

    def __str__(self):
        sql_short = self.sql.strip()[0:40].replace("\n", "").strip()
        return f"<Raw SQL: {self.name}/{self.stage}:{sql_short}>"

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if value is not None and not isinstance(value, str):
            raise TypeError(f"Table name must be of instance 'str' not {type(value)}.")
        self._name = normalize_name(value)


class Blob(Generic[T]):
    """Blob (binary large object) container

    Used to wrap arbitrary python objects that get returned from materializing
    tasks.

    :param obj: The object to wrap
    :param name: Optional name. If no name is provided, an automatically
        generated name will be used. To prevent name collisions, you can
        add '%%' at the end of the name to enable automatic name mangling.
    :param cache_key: Internal cache_key used when retreiving an object
        from the database cache.
    """

    def __init__(
        self,
        obj: T | None = None,
        name: str | None = None,
        stage: Stage | None = None,
    ):
        self._name = None

        self.obj = obj
        self.name = name
        self.stage = stage

        # cache_key will be overridden shortly before handing over to downstream tasks
        # that use it to compute their input_hash for cache_invalidation due to input
        # change
        self.cache_key = None

    def __str__(self):
        return f"<Blob: {self.name} ({self.stage.name})>"

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = normalize_name(value)



from __future__ import annotations

from contextvars import ContextVar, Token
from enum import Enum
from functools import cached_property
from threading import Lock
from typing import TYPE_CHECKING, Any, ClassVar

from pydiverse.pipedag.util.config import load_object
from pydiverse.pipedag.util.import_ import import_object

if TYPE_CHECKING:
    from pydiverse.pipedag._typing import T
    from pydiverse.pipedag.backend import BaseLockManager
    from pydiverse.pipedag.context.run_context import StageLockStateHandler
    from pydiverse.pipedag.core import Flow, Stage, Task
    from pydiverse.pipedag.engine.base import OrchestrationEngine
    from pydiverse.pipedag.materialize.metadata import TaskMetadata

import structlog
from attrs import define, frozen

logger = structlog.get_logger()


class BaseContext:
    _context_var: ClassVar[ContextVar]
    _token: Token = None
    _enter_counter: int = 0
    _lock: Lock = Lock()

    def __enter__(self):
        with self._lock:
            object.__setattr__(self, "_enter_counter", self._enter_counter + 1)

            if self._token is not None:
                return self

            token = self._context_var.set(self)
            object.__setattr__(self, "_token", token)
        return self

    def __exit__(self, *_):
        with self._lock:
            object.__setattr__(self, "_enter_counter", self._enter_counter - 1)
            if self._enter_counter == 0:
                if not self._token:
                    raise RuntimeError
                self._context_var.reset(self._token)
                object.__setattr__(self, "_token", None)
                self.close()

    def close(self):
        """Function that gets called at __exit__"""

    @classmethod
    def get(cls: type[T]) -> T:
        return cls._context_var.get()

    def __getstate__(self):
        state = self.__dict__.copy()
        del state["_token"]
        del state["_enter_counter"]
        return state


@frozen(slots=False)
class BaseAttrsContext(BaseContext):
    pass


@frozen
class DAGContext(BaseAttrsContext):
    """Context during DAG definition"""

    flow: Flow
    stage: Stage

    _context_var = ContextVar("dag_context")


@define
class TaskContext(BaseContext):
    """Context used while executing a task

    It is used to retrieve a reference to the task object from within a running task.
    It also serves as a place to store temporary state while processing a task.
    """

    task: Task

    # Temporary state associated with a task during materialization
    input_hash: str = None
    cache_fn_hash: str = None
    cache_metadata: TaskMetadata = None

    _context_var = ContextVar("task_context")


class StageCommitTechnique(Enum):
    SCHEMA_SWAP = 0
    READ_VIEWS = 1


@frozen(slots=False)
class ConfigContext(BaseAttrsContext):
    """
    Configuration context for running a particular pipedag instance.

    For the most part it holds serializable attributes. But it also offers access
    to blob_store and table_store as well as lock manager and orchestration engine.
    Lock manager and orchestration engine are managed full cycle by exactly one
    caller (ServerRunContext / Flow.run()). So this class just offers a way to create
    a disposable object. Blob_store and table_store on the other hand might be
    accessed by multi-threaded / multi-processed / multi-node way of orchestrating
    functions in the pipedag. Since they don't keep real state, they can be
    thrown away while pickling and will be reloaded on first access of the
    @cached_property store. Calling ConfigContext.dispose() will close all
    connections and render this object unusable afterwards.
    """

    config_dict: dict

    # names
    pipedag_name: str
    flow_name: str
    instance_name: str

    # per instance attributes
    fail_fast: bool
    strict_result_get_locking: bool
    instance_id: str  # may be used as database name or locking ID
    stage_commit_technique: StageCommitTechnique
    network_interface: str
    attrs: dict[str, Any]

    @cached_property
    def auto_table(self) -> tuple[type, ...]:
        return tuple(map(import_object, self.config_dict.get("auto_table", ())))

    @cached_property
    def auto_blob(self) -> tuple[type, ...]:
        return tuple(map(import_object, self.config_dict.get("auto_blob", ())))

    @cached_property
    def store(self):
        # Load objects referenced in config
        table_store = load_object(self.config_dict["table_store"])
        blob_store = load_object(self.config_dict["blob_store"])
        from pydiverse.pipedag.materialize.store import PipeDAGStore

        return PipeDAGStore(
            table=table_store,
            blob=blob_store,
        )

    def create_lock_manager(self) -> BaseLockManager:
        return load_object(self.config_dict["lock_manager"])

    def create_orchestration_engine(self) -> OrchestrationEngine:
        return load_object(self.config_dict["orchestration"])

    def close(self):
        # If the store has been initialized (and thus cached in the __dict__),
        # dispose of it, and remove it from the cache.
        if store := self.__dict__.get("store", None):
            store.dispose()
            self.__dict__.pop("store")

    def __getstate__(self):
        state = super().__getstate__()
        # store is not serializable, but @cached_property will reload
        # it from config_dict
        state.pop("store", None)
        state.pop("auto_table", None)
        state.pop("auto_blob", None)
        return state

    _context_var = ContextVar("config_context")


class StageLockContext(BaseContext):
    """
    Context manager used to keep stages locked until after flow.run() has been called
    """

    lock_state_handlers: [StageLockStateHandler]

    _context_var = ContextVar("stage_lock_context")

    def __init__(self):
        self.logger = structlog.getLogger(module=__name__, cls=self.__class__.__name__)
        self.logger.info(f"OPEN STAGE LOCK CONTEXT {id(self)}")
        self.lock_state_handlers = []

    def close(self):
        self.logger.info(f"CLOSE STAGE LOCK CONTEXT {id(self)}")
        for lock_state_handler in self.lock_state_handlers:
            lock_state_handler.dispose()



from __future__ import annotations

import functools
import inspect
from functools import partial
from typing import TYPE_CHECKING, Callable

from pydiverse.pipedag._typing import CallableT
from pydiverse.pipedag.context import ConfigContext, RunContext, TaskContext
from pydiverse.pipedag.core.task import Task
from pydiverse.pipedag.materialize.cache import CacheManager
from pydiverse.pipedag.materialize.container import Blob, Table
from pydiverse.pipedag.materialize.util import compute_cache_key
from pydiverse.pipedag.util import deep_map

if TYPE_CHECKING:
    pass


def materialize(
    fn: CallableT = None,
    *,
    name: str = None,
    input_type: type = None,
    version: str = None,
    cache: Callable = None,
    lazy: bool = False,
    nout: int = 1,
) -> CallableT | MaterializingTask:
    if fn is None:
        return partial(
            materialize,
            name=name,
            input_type=input_type,
            version=version,
            cache=cache,
            lazy=lazy,
            nout=nout,
        )

    return MaterializingTask(
        fn,
        name=name,
        input_type=input_type,
        version=version,
        cache=cache,
        lazy=lazy,
        nout=nout,
    )


class MaterializingTask(Task):
    """Task whose outputs get materialized

    All the values a materializing task returns get written to the appropriate
    storage backend. Additionally, all `Table` and `Blob` objects in the
    input will be replaced with their appropriate objects (loaded from the
    storage backend). This means that tables and blobs never move from one
    task to another directly, but go through the storage layer instead.

    Because of how caching is implemented, task inputs and outputs must all
    be 'materializable'. This means that they can only contain objects of
    the following types:
    `dict`, `list`, `tuple`,
    `int`, `float`, `str`, `bool`, `None`,
    and PipeDAG's `Table` and `Blob` type.

    Automatically adds itself to the active stage.
    All materializing tasks MUST be defined inside a stage.

    :param fn: The run method of this task
    :key name: The name of this task
    :key input_type: The data type to convert table objects to when passed
        to this task.
    :key version: The version of this task. Unless this task is lazy, you
        always have to bump / change the version number to ensure that
        the new implementation gets used. Else a cached result might be used
        instead.
    :key cache: An explicit function for validating cache validity. If the output
        of this function changes while the source parameters are the same (e.g.
        the source is a filepath and `cache` loads data from this file), then the
        cache will be deemed invalid and is not used.
    :key lazy: Boolean indicating if this task should be lazy. A lazy task is
        a task that always gets executed, and if it produces a lazy table
        (e.g. a SQL query), the backend can compare the generated output
        to see it the same query has been executed before (and only execute
        it if not). This is an alternative to manually setting the version
        number.
    :key kwargs: Any other keyword arguments will directly get passed to the
        prefect Task initializer.
    """

    def __init__(
        self,
        fn: Callable,
        *,
        name: str = None,
        input_type: type = None,
        version: str = None,
        cache: Callable = None,
        lazy: bool = False,
        nout: int = 1,
    ):
        super().__init__(
            MaterializationWrapper(fn),
            name=name,
            nout=nout,
        )

        self.input_type = input_type
        self.version = version
        self.cache = cache
        self.lazy = lazy


class MaterializationWrapper:
    """Function wrapper that contains all high level materialization logic

    :param fn: The function to wrap
    """

    def __init__(self, fn: Callable):
        functools.update_wrapper(self, fn)

        self.fn = fn
        self.fn_signature = inspect.signature(fn)

    def __call__(self, *args, **kwargs):
        """Function wrapper / materialization logic

        :param args: The arguments passed to the function
        :param _pipedag_task_: The `MaterializingTask` instance which called
            this wrapper.
        :param kwargs: The keyword arguments passed to the function
        :return: A copy of what the original function returns annotated
            with some additional metadata.
        """

        task: MaterializingTask = TaskContext.get().task  # type: ignore
        store = ConfigContext.get().store
        bound = self.fn_signature.bind(*args, **kwargs)

        if task is None:
            raise TypeError("Task can't be None.")

        # If this is the first task in this stage to be executed, ensure that
        # the stage has been initialized and locked.
        store.ensure_stage_is_ready(task.stage)

        # Compute the cache key for the task inputs
        input_json = store.json_encode(bound.arguments)
        input_hash = compute_cache_key("INPUT", input_json)

        cache_fn_hash = ""
        if task.cache is not None:
            cache_fn_output = store.json_encode(task.cache(*args, **kwargs))
            cache_fn_hash = compute_cache_key("CACHE_FN", cache_fn_output)

        memo_cache_key = CacheManager.task_cache_key(task, input_hash, cache_fn_hash)

        # Check if this task has already been run with the same inputs
        # If yes, return memoized result. This prevents DuplicateNameExceptions
        ctx = RunContext.get()
        with ctx.task_memo(task, memo_cache_key) as (success, memo):
            if success:
                task.logger.info(
                    "Task has already been run with the same inputs."
                    " Using memoized results."
                )

                return memo

            task_cache_info = CacheManager.cache_lookup(
                store, task, input_hash, cache_fn_hash
            )
            if task_cache_info.is_cache_valid():
                if not task.lazy:
                    ctx = RunContext.get()
                    ctx.store_task_memo(
                        task, memo_cache_key, task_cache_info.get_cached_output()
                    )
                    # Task isn't lazy -> copy cache to transaction stage
                    return task_cache_info.get_cached_output()

            # Not found in cache / lazy -> Evaluate Function
            args, kwargs = store.dematerialize_task_inputs(
                task, bound.args, bound.kwargs
            )

            result = self.fn(*args, **kwargs)
            result = store.materialize_task(task, task_cache_info, result)

            # Delete underlying objects from result (after materializing them)
            def obj_del_mutator(x):
                if isinstance(x, (Table, Blob)):
                    x.obj = None
                return x

            result = deep_map(result, obj_del_mutator)
            ctx.store_task_memo(task, memo_cache_key, result)
            self.value = result

            return result



from pydiverse.pipedag.context.context import (
    ConfigContext,
    DAGContext,
    StageLockContext,
    TaskContext,
)
from pydiverse.pipedag.context.run_context import RunContext, RunContextServer

__all__ = [
    "DAGContext",
    "TaskContext",
    "ConfigContext",
    "RunContext",
    "RunContextServer",
    "StageLockContext",
]



from __future__ import annotations

import itertools
import json
from typing import Any, Callable

import structlog

from pydiverse.pipedag import Blob, Stage, Table, backend
from pydiverse.pipedag._typing import Materializable
from pydiverse.pipedag.context import ConfigContext, RunContext
from pydiverse.pipedag.context.run_context import StageState
from pydiverse.pipedag.errors import DuplicateNameError, StageError
from pydiverse.pipedag.materialize.cache import TaskCacheInfo
from pydiverse.pipedag.materialize.container import RawSql
from pydiverse.pipedag.materialize.core import MaterializingTask
from pydiverse.pipedag.materialize.metadata import TaskMetadata
from pydiverse.pipedag.materialize.util import json as json_util
from pydiverse.pipedag.util import Disposable, deep_map
from pydiverse.pipedag.util.config import PipedagConfig


class PipeDAGStore(Disposable):
    """Main storage interface for materializing tasks

    Depending on the use case, the store can be configured using different
    backends for storing tables, blobs and managing locks.

    Other than initializing the global `PipeDAGStore` object, the user
    should never have to interact with it. It only serves as a coordinator
    between the different backends, the stages and the materializing tasks.
    """

    def __init__(
        self,
        table: backend.table.BaseTableStore,
        blob: backend.blob.BaseBlobStore,
    ):
        self.table_store = table
        self.blob_store = blob

        self.logger = structlog.get_logger()
        self.json_encoder = json.JSONEncoder(
            ensure_ascii=False,
            allow_nan=False,
            separators=(",", ":"),
            sort_keys=True,
            default=json_util.json_default,
        )
        self.json_decoder = json.JSONDecoder(object_hook=json_util.json_object_hook)

    def dispose(self):
        """
        Clean up and close all open resources.

        Don't use the store object any more after disposal!
        """
        self.table_store.dispose()
        self.blob_store.dispose()
        super().dispose()

    # ### Stage ### #

    def init_stage(self, stage: Stage):
        """Initializes the stage in all backends

        This function also acquires a lock on the given stage to prevent
        other flows from modifying the same stage at the same time. Only
        once all tasks that depend on this stage have been executed, will
        this lock be released.

        Don't use this function directly. Instead, use `ensure_stage_is_ready`
        to prevent unnecessary locking.
        """

        RunContext.get().validate_stage_lock(stage)
        self.table_store.init_stage(stage)
        self.blob_store.init_stage(stage)

        # Once stage reference counter hits 0 (this means all tasks that
        # have any task contained in the stage as an upstream dependency),
        # the lock gets automatically released by the RunContextServer

    def ensure_stage_is_ready(self, stage: Stage):
        """Initializes a stage if it hasn't been created yet

        This allows the creation of stages in a lazy way. This ensures
        that a stage only gets locked right before it is needed.
        """
        with RunContext.get().init_stage(stage) as should_continue:
            if not should_continue:
                return
            self.init_stage(stage)

    def commit_stage(self, stage: Stage):
        """Commit the stage"""
        ctx = RunContext.get()
        with ctx.commit_stage(stage) as should_continue:
            if not should_continue:
                raise StageError
            ctx.validate_stage_lock(stage)
            self.table_store.commit_stage(stage)
            self.blob_store.commit_stage(stage)

    # ### Materialization ### #

    def dematerialize_item(
        self, item: Table | Blob | Any, as_type: type, ctx: RunContext = None
    ):
        if ctx is None:
            ctx = RunContext.get()

        if isinstance(item, Table):
            ctx.validate_stage_lock(item.stage)
            return self.table_store.retrieve_table_obj(item, as_type=as_type)
        elif isinstance(item, Blob):
            ctx.validate_stage_lock(item.stage)
            return self.blob_store.retrieve_blob(item)
        return item

    def dematerialize_task_inputs(
        self,
        task: MaterializingTask,
        args: tuple[Materializable],
        kwargs: dict[str, Materializable],
    ) -> tuple[tuple, dict]:
        """Loads the inputs for a task from the storage backends

        Traverses the function arguments and replaces all `Table` and
        `Blob` objects with the associated objects stored in the backend.

        :param task: The task for which the arguments should be dematerialized
        :param args: The positional arguments
        :param kwargs: The keyword arguments
        :return: A tuple with the dematerialized args and kwargs
        """

        ctx = RunContext.get()

        def dematerialize_mapper(x):
            return self.dematerialize_item(x, as_type=task.input_type, ctx=ctx)

        d_args = deep_map(args, dematerialize_mapper)
        d_kwargs = deep_map(kwargs, dematerialize_mapper)

        return d_args, d_kwargs

    def materialize_task(
        self,
        task: MaterializingTask,
        task_cache_info: TaskCacheInfo,
        value: Materializable,
    ) -> Materializable:
        """Stores the output of a task in the backend

        Traverses the output produced by a task, adds missing metadata,
        materializes all `Table` and `Blob` objects and returns a new
        output object with the required metadata to allow dematerialization.

        :param task: The task instance which produced `value`. Must have
            the correct `cache_key` attribute set.
        :param task_cache_info: Information about task carried through materialization
            for CacheManager
        :param value: The output of the task. Must be materializable; this
            means it can only contain the following object types:
            `dict`, `list`, `tuple`,
            `int`, `float`, `str`, `bool`, `None`,
            and PipeDAG's `Table` and `Blob` type.
        :return: A copy of `value` with additional metadata
        """

        stage = task.stage
        ctx = RunContext.get()

        if (state := ctx.get_stage_state(stage)) != StageState.READY:
            raise StageError(
                f"Can't materialize because stage '{stage.name}' is not ready "
                f"(state: {state})."
            )

        tables = []
        raw_sqls = []
        blobs = []

        config = ConfigContext.get()
        auto_suffix_counter = itertools.count()

        def materialize_mutator(x):
            # Automatically convert an object to a table / blob if its
            # type is inside either `config.auto_table` or `.auto_blob`.
            if isinstance(x, config.auto_table):
                try:
                    hook = self.table_store.get_m_table_hook(type(x))
                    x = hook.auto_table(x)
                except TypeError:
                    x = Table(x)
            if isinstance(x, config.auto_blob):
                x = Blob(x)

            if isinstance(x, PipedagConfig):
                # Config objects are not an allowed return type,
                # because they might mess up caching.
                raise TypeError(
                    "You can't return a PipedagConfig object from a materializing task."
                )

            # Do the materialization
            if isinstance(x, (Table, RawSql, Blob)):
                if not task.lazy:
                    # task cache_key is output cache_key for eager tables
                    x.cache_key = task_cache_info.get_task_cache_key()

                if isinstance(x, (Table, Blob)):
                    x.stage = stage
                    # Update name:
                    # - If no name has been provided, generate on automatically
                    # - If the provided name ends with %%, perform name mangling
                    object_number = next(auto_suffix_counter)
                    auto_suffix = (
                        f"{task_cache_info.get_task_cache_key()}_{object_number:04d}"
                    )
                    if x.name is None:
                        x.name = task.name + "_" + auto_suffix
                    elif x.name.endswith("%%"):
                        x.name = x.name[:-2] + auto_suffix

                ctx.validate_stage_lock(stage)
                if isinstance(x, Table):
                    if x.obj is None:
                        raise TypeError("Underlying table object can't be None")
                    tables.append(x)
                elif isinstance(x, RawSql):
                    if x.sql is None:
                        raise TypeError("Underlying raw sql string can't be None")
                    raw_sqls.append(x)
                elif isinstance(x, Blob):
                    if task.lazy:
                        raise NotImplementedError(
                            "Can't use Blobs with lazy tasks. Invalidation of the"
                            " downstream dependencies is not implemented."
                        )
                    blobs.append(x)
                else:
                    raise NotImplementedError

            return x

        m_value = deep_map(value, materialize_mutator)

        def store_metadata():
            """
            Metadata must be generated after everything else has been materialized,
            because during materialization the cache_key of the different objects
            can get changed.
            """
            output_json = self.json_encode(m_value)
            task_cache_info.store_task_metadata(output_json, self.table_store, stage)

        def store_table(table: Table):
            if task.lazy:
                self.table_store.store_table_lazy(table, task, task_cache_info)
            else:
                self.table_store.store_table(table)

        # Materialize
        self._check_names(task, tables, blobs)
        self._store_task_transaction(
            task,
            tables,
            raw_sqls,
            blobs,
            store_table,
            lambda raw_sql: self.table_store.store_raw_sql(
                raw_sql, task, task_cache_info
            ),
            self.blob_store.store_blob,
            store_metadata,
        )

        return m_value

    @staticmethod
    def _check_names(task: MaterializingTask, tables: list[Table], blobs: list[Blob]):
        if not tables and not blobs:
            # Nothing to check
            return

        # Check names: No duplicates in task
        seen_tn = set()
        seen_bn = set()
        tn_dup = [e.name for e in tables if e.name in seen_tn or seen_tn.add(e.name)]
        bn_dup = [e.name for e in blobs if e.name in seen_bn or seen_bn.add(e.name)]

        if tn_dup or bn_dup:
            raise DuplicateNameError(
                f"Task '{task.name}' returned multiple tables and/or blobs"
                " with the same name.\n"
                f"Duplicate table names: {', '.join(tn_dup) if tn_dup else 'None'}\n"
                f"Duplicate blob names: {', '.join(bn_dup) if bn_dup else 'None'}\n"
                "To enable automatic name mangling,"
                " you can add '%%' at the end of the name."
            )

        # Check names: No duplicates in stage
        ctx = RunContext.get()
        success, tn_dup, bn_dup = ctx.add_names(task.stage, tables, blobs)

        if not success:
            raise DuplicateNameError(
                f"Task '{task.name}' returned tables and/or blobs"
                f" whose name are not unique in the schema '{task.stage}'.\n"
                f"Duplicate table names: {', '.join(tn_dup) if tn_dup else 'None'}\n"
                f"Duplicate blob names: {', '.join(bn_dup) if bn_dup else 'None'}\n"
                "To enable automatic name mangling,"
                " you can add '%%' at the end of the name."
            )

    def _store_task_transaction(
        self,
        task: MaterializingTask,
        tables: list[Table],
        raw_sqls: list[RawSql],
        blobs: list[Blob],
        store_table: Callable[[Table], None],
        store_raw_sql: Callable[[RawSql], None],
        store_blob: Callable[[Blob], None],
        store_metadata: Callable[[], None],
    ):
        stage = task.stage
        ctx = RunContext.get()

        stored_tables = []
        stored_blobs = []

        try:
            for table in tables:
                ctx.validate_stage_lock(stage)
                store_table(table)
                stored_tables.append(table)
            for blob in blobs:
                ctx.validate_stage_lock(stage)
                store_blob(blob)
                stored_blobs.append(blob)
            for raw_sql in raw_sqls:
                ctx.validate_stage_lock(stage)
                store_raw_sql(raw_sql)

            ctx.validate_stage_lock(task.stage)
            store_metadata()

        except Exception as e:
            # Failed - Roll back everything
            for table in stored_tables:
                self.table_store.delete_table_from_transaction(table)
            for blob in stored_blobs:
                self.blob_store.delete_blob_from_transaction(blob)

            ctx.remove_names(stage, tables, blobs)
            raise e

    # ### Cache ### #

    def retrieve_cached_output(
        self,
        task: MaterializingTask,
        input_hash: str,
        cache_fn_hash: str,
    ) -> (Materializable, TaskMetadata):
        """Try to retrieve the cached outputs for a task

        :param task: The materializing task for which to retrieve
            the cached output. Must have the `cache_key` attribute set.
        :raises CacheError: if no matching task exists in the cache
        """

        if task.stage.did_commit:
            raise StageError(f"Stage ({task.stage}) already committed.")

        metadata = self.table_store.retrieve_task_metadata(
            task, input_hash, cache_fn_hash
        )
        return self.json_decode(metadata.output_json), metadata

    def copy_cached_output_to_transaction_stage(
        self,
        output: Materializable,
        original_metadata: TaskMetadata,
        task: MaterializingTask,
    ):
        """Copy the (non-lazy) outputs from a cached task into the transaction stage

        If the outputs of a task were successfully retrieved from the cache
        using `retrieve_cached_output`, they and the associated metadata
        must be copied from the base stage to the transaction.

        :raises CacheError: if some values in the output can't be found
            in the cache.
        """

        # Get Tables and Blobs from output
        tables = []
        blobs = []
        raw_sqls = []

        def visitor(x):
            if isinstance(x, Table):
                tables.append(x)
            elif isinstance(x, RawSql):
                raw_sqls.append(x)
            elif isinstance(x, Blob):
                blobs.append(x)
            return x

        deep_map(output, visitor)

        def store_raw_sql(raw_sql):
            raise Exception("raw sql scripts cannot be part of a non-lazy task")

        # Materialize
        self._check_names(task, tables, blobs)
        self._store_task_transaction(
            task,
            tables,
            raw_sqls,
            blobs,
            self.table_store.copy_table_to_transaction,
            store_raw_sql,
            self.blob_store.copy_blob_to_transaction,
            lambda: self.table_store.store_task_metadata(original_metadata, task.stage),
        )

    # ### Utils ### #

    def json_encode(self, value: Materializable) -> str:
        """Encode a materializable value as json

        In addition to the default types that python can serialise to json,
        this function can also serialise `Table`, `RawSql`, and `Blob` objects.
        The only caveat is, that python's json module doesn't differentiate
        between lists and tuples, which means it is impossible to
        differentiate the two.
        """
        return self.json_encoder.encode(value)

    def json_decode(self, value: str) -> Materializable:
        """Decode a materializable value as json

        Counterpart for the `json_encode` function. Can decode `Table` and
        `Blob` objects.
        """
        return self.json_decoder.decode(value)



from .config import PipedagConfig, setup_structlog
from .deep_map import deep_map
from .deep_merge import deep_merge
from .disposable import Disposable
from .import_ import requires
from .naming import normalize_name, safe_name



from __future__ import annotations

import ipaddress
import os
import struct
import threading
import uuid
from functools import cached_property
from typing import Any

import msgpack
import pynng
import structlog

from pydiverse.pipedag.errors import IPCError


def human_thread_id(thread_ident):
    """Return human processable number for thread ID."""
    return hash(thread_ident) % 100


class IPCServer(threading.Thread):
    """Server for inter process communication

    FORMAT:
        REQUEST: [nonce, payload]
        RESPONSE: [payload]
    """

    def __init__(
        self,
        listen="tcp://127.0.0.1:0",
        msg_default=None,
        msg_ext_hook=None,
    ):
        super().__init__(name="IPCServer", daemon=True)

        # Reduce recv_timeout when running pytest for better performance
        is_running_pytest = "PYDIVERSE_PIPEDAG_PYTEST" in os.environ
        recv_timeout = 200 if not is_running_pytest else 10

        self.socket = pynng.Rep0(listen=listen, recv_timeout=recv_timeout)
        self.nonces = set()

        self.__stop_flag = False
        self.__thread = None
        self.logger = structlog.get_logger(cls=type(self).__name__)

        self.msg_default = msg_default
        self.msg_ext_hook = msg_ext_hook

    def _get_id(self, ctx_id=0):
        return dict(
            address=self.address,
            thread=human_thread_id(threading.get_ident()),
            ctx_id=ctx_id,
        )

    def run(self):
        self.logger.info("start IPCServer", **self._get_id())
        self.run_loop()

    def stop(self):
        self.__stop_flag = True
        self.logger.debug("request IPCServer to stop", **self._get_id())

    def __enter__(self):
        self.__thread = self
        self.__thread.start()
        self.logger.debug(
            "started IPCServer thread", thread=human_thread_id(self.__thread.ident)
        )
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()
        if self.__thread is not None:
            self.logger.debug(
                "stop IPCServer thread", thread=human_thread_id(self.__thread.ident)
            )
            self.__thread.join()

    def run_loop(self):
        """
        Run main loop for responding to client requests to this server process.

        We rely on a fixed request-response pattern of pynng library (Req0/Rep0)
        which requires us to open a context for every request-response pair.
        We create threads for determining and sending responses and we keep 10
        such threads in-flight. At the end we join all 10 threads even though
        some threads might actually already have completed and may even share the
        same ident number. We log thread IDs as hash(thread.ident) % 100 for
        receiving human processable numbers.
        """
        max_threads_in_flight = 10  # keep up to 10 threads in flight

        socket = None
        ctx_id = 0
        threads = []
        try:
            socket = self.socket.new_context()
            ctx_id += 1
            self.logger.debug("new socket context", **self._get_id(ctx_id))
            while not self.__stop_flag:
                # noinspection PyBroadException
                try:
                    data = socket.recv()
                    unpacker = msgpack.Unpacker(
                        use_list=False, ext_hook=self.msg_ext_hook
                    )
                    unpacker.feed(data)

                    # Expected: (NONCE, PAYLOAD)
                    if unpacker.read_array_header() != 2:
                        raise IPCError(
                            "Expected request to contain exactly two fields:"
                            " nonce and payload."
                        )
                    nonce = unpacker.unpack()
                    nonce_hex = nonce.hex()

                    if nonce in self.nonces:
                        # Already processing this request
                        self.logger.warning(
                            "Already processing request",
                            nonce=nonce_hex,
                            **self._get_id(ctx_id),
                        )
                        socket = self.socket.new_context()
                        ctx_id += 1
                        self.logger.debug("next socket context", **self._get_id(ctx_id))
                        continue

                    self.nonces.add(nonce)

                    if len(threads) >= max_threads_in_flight:
                        self.logger.debug(
                            "joining thread", thread=human_thread_id(threads[0].ident)
                        )
                        threads[0].join()
                        del threads[0]
                    thread_logger = structlog.get_logger()
                    thread = threading.Thread(
                        name="IPC Worker",
                        target=self._serve,
                        args=[thread_logger, socket, unpacker, nonce_hex, ctx_id],
                        daemon=True,
                    )
                    socket = self.socket.new_context()
                    ctx_id += 1
                    self.logger.debug("next socket context", **self._get_id(ctx_id))
                    threads.append(thread)
                    thread.start()
                except pynng.Timeout:
                    pass
                except Exception:
                    self.logger.exception("exception occurred in run_loop")
        finally:
            for thread in threads:
                self.logger.debug(
                    "joining thread", thread=human_thread_id(thread.ident)
                )
                thread.join()
            self.logger.debug("closing socket", **self._get_id())
            if socket is not None:
                socket.close()  # close the open request-response context
            self.socket.close()
            self.logger.info("stopped IPCServer", **self._get_id())

    def _serve(self, thread_logger, socket, unpacker, nonce_hex, ctx_id):
        msg = unpacker.unpack()
        thread_logger.debug(
            "IPCServer Received",
            message=msg,
            nonce=nonce_hex,
            ctx_id=ctx_id,
            thread=human_thread_id(threading.get_ident()),
        )
        reply = self.handle_request(msg)
        thread_logger.debug(
            "IPCServer Reply",
            reply=reply,
            nonce=nonce_hex,
            ctx_id=ctx_id,
            thread=human_thread_id(threading.get_ident()),
        )
        socket.send(msgpack.packb(reply, default=self.msg_default))

    def handle_request(self, request: dict[str, Any]):
        return None

    @cached_property
    def address(self) -> str:
        address = self.socket.listeners[0].local_address
        addr = address.addr
        port = address.port

        # Convert from big endian to native
        if isinstance(addr, int):
            addr = struct.unpack("=L", addr.to_bytes(4, byteorder="big"))[0]
        port = struct.unpack("=H", port.to_bytes(2, byteorder="big"))[0]

        # Create address string
        ip = ipaddress.ip_address(addr)
        if ip.version == 4:
            return f"tcp://{ip.exploded}:{port}"
        else:
            return f"tcp://[{ip}]:{port}"

    def get_client(self) -> IPCClient:
        return IPCClient(
            addr=self.address,
            msg_default=self.msg_default,
            msg_ext_hook=self.msg_ext_hook,
        )


class IPCClient:
    def __init__(self, addr: str, msg_default=None, msg_ext_hook=None):
        self.logger = structlog.get_logger(
            thread=human_thread_id(threading.get_ident())
        )
        self.addr = addr
        self.socket = self._connect()

        self.msg_default = msg_default
        self.msg_ext_hook = msg_ext_hook

    def _connect(self):
        self.logger.debug("opening client connection", addr=self.addr)
        return pynng.Req0(dial=self.addr, resend_time=30_000)

    def request(self, payload: Any) -> Any:
        with self.socket.new_context() as socket:
            self.logger.debug("client request")
            nonce = uuid.uuid4().bytes[:8]
            msg = msgpack.packb((nonce, payload), default=self.msg_default)
            socket.send(msg)

            response_msg = socket.recv()
            self.logger.debug("client got response")
            response = msgpack.unpackb(
                response_msg, use_list=False, ext_hook=self.msg_ext_hook
            )
            return response

    def __getstate__(self):
        state = self.__dict__.copy()
        del state["socket"]
        del state["logger"]
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.logger = structlog.get_logger(threading.get_ident())
        self.socket = self._connect()


if __name__ == "__main__":
    with IPCServer() as serv:
        client = serv.get_client()
        client.request(["This is a test"])



from __future__ import annotations

import copy
import re

import sqlalchemy as sa
from attr import frozen
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.schema import DDLElement
from sqlalchemy.sql import Select
from sqlalchemy.sql.elements import TextClause

__all__ = [
    "Schema",
    "CreateSchema",
    "DropSchema",
    "RenameSchema",
    "CreateTableAsSelect",
    "CreateViewAsSelect",
    "CopyTable",
    "DropTable",
    "CreateDatabase",
    "DropFunction",
    "DropProcedure",
    "DropView",
    "AddPrimaryKey",
    "AddIndex",
    "ChangeColumnNullable",
    "ChangeColumnTypes",
    "split_ddl_statement",
]


@frozen
class Schema:
    name: str
    prefix: str
    suffix: str

    def get(self):
        return self.prefix + self.name + self.suffix


class CreateSchema(DDLElement):
    def __init__(self, schema: Schema, if_not_exists=False):
        self.schema = schema
        self.if_not_exists = if_not_exists


class DropSchema(DDLElement):
    def __init__(self, schema: Schema, if_exists=False, cascade=False, *, engine=None):
        """
        :param engine: Used if cascade=True but the database doesn't support cascade.
        """
        self.schema = schema
        self.if_exists = if_exists
        self.cascade = cascade

        self.engine = engine


class RenameSchema(DDLElement):
    def __init__(self, from_: Schema, to: Schema):
        self.from_ = from_
        self.to = to


class CreateDatabase(DDLElement):
    def __init__(self, database: str, if_not_exists=False):
        self.database = database
        self.if_not_exists = if_not_exists


class DropDatabase(DDLElement):
    def __init__(self, database: str, if_exists=False, cascade=False):
        self.database = database
        self.if_exists = if_exists
        self.cascade = cascade


class CreateTableAsSelect(DDLElement):
    def __init__(self, name: str, schema: Schema, query: Select | TextClause | sa.Text):
        self.name = name
        self.schema = schema
        self.query = query


class CreateViewAsSelect(DDLElement):
    def __init__(self, name: str, schema: Schema, query: Select | TextClause | sa.Text):
        self.name = name
        self.schema = schema
        self.query = query


class CopyTable(DDLElement):
    def __init__(
        self,
        from_name,
        from_schema: Schema,
        to_name,
        to_schema: Schema,
        if_not_exists=False,
    ):
        self.from_name = from_name
        self.from_schema = from_schema
        self.to_name = to_name
        self.to_schema = to_schema
        self.if_not_exists = if_not_exists


class DropTable(DDLElement):
    def __init__(self, name, schema: Schema, if_exists=False):
        self.name = name
        self.schema = schema
        self.if_exists = if_exists


class DropView(DDLElement):
    """
    Attention: For mssql, this statement must be prefixed with
               a 'USE <database>' statement.
    """

    def __init__(self, name, schema: Schema, if_exists=False):
        self.name = name
        self.schema = schema
        self.if_exists = if_exists


class DropProcedure(DDLElement):
    """
    Attention: For mssql, this statement must be prefixed with
               a 'USE <database>' statement.
    """

    def __init__(self, name, schema: Schema, if_exists=False):
        self.name = name
        self.schema = schema
        self.if_exists = if_exists


class DropFunction(DDLElement):
    """
    Attention: For mssql, this statement must be prefixed with
               a 'USE <database>' statement.
    """

    def __init__(self, name, schema: Schema, if_exists=False):
        self.name = name
        self.schema = schema
        self.if_exists = if_exists


class AddPrimaryKey(DDLElement):
    def __init__(
        self,
        table_name: str,
        schema: Schema,
        key_columns: list[str],
        name: str | None = None,
    ):
        self.table_name = table_name
        self.schema = schema
        self.key = key_columns
        self.name = name


class AddIndex(DDLElement):
    def __init__(
        self,
        table_name: str,
        schema: Schema,
        index_columns: list[str],
        name: str | None = None,
    ):
        self.table_name = table_name
        self.schema = schema
        self.index = index_columns
        self.name = name


class ChangeColumnTypes(DDLElement):
    def __init__(
        self,
        table_name: str,
        schema: Schema,
        column_names: list[str],
        column_types: list[str],
        nullable: bool | list[bool] | None = None,
        cap_varchar_max: int | None = None,
    ):
        if not isinstance(nullable, list):
            nullable = [nullable for _ in column_names]
        self.table_name = table_name
        self.schema = schema
        self.column_names = column_names
        self.column_types = column_types
        self.nullable = nullable
        self.cap_varchar_max = cap_varchar_max


class ChangeColumnNullable(DDLElement):
    def __init__(
        self,
        table_name: str,
        schema: Schema,
        column_names: list[str],
        nullable: bool | list[bool],
    ):
        if isinstance(nullable, bool):
            nullable = [nullable for _ in column_names]
        self.table_name = table_name
        self.schema = schema
        self.column_names = column_names
        self.nullable = nullable


@compiles(CreateSchema)
def visit_create_schema(create: CreateSchema, compiler, **kw):
    _ = kw
    schema = compiler.preparer.format_schema(create.schema.get())
    text = ["CREATE SCHEMA"]
    if create.if_not_exists:
        text.append("IF NOT EXISTS")
    text.append(schema)
    return " ".join(text)


# noinspection SqlDialectInspection
@compiles(CreateSchema, "mssql")
def visit_create_schema_mssql(create: CreateSchema, compiler, **kw):
    # For SQL Server we support two modes:  using databases as schemas,
    # or schemas as schemas.
    _ = kw
    if "." in create.schema.name:
        raise AttributeError(
            "We currently do not support dots in schema names "
            " when working with mssql database"
        )
    full_name = create.schema.get()
    # it was already checked that there is exactly one dot in schema prefix + suffix
    database_name, schema_name = full_name.split(".")
    database = compiler.preparer.format_schema(database_name)
    schema = compiler.preparer.format_schema(schema_name)
    create_schema = f"CREATE SCHEMA {schema}"
    create_database = f"CREATE DATABASE {database}"
    if create.if_not_exists:
        create_database = f"""
            IF NOT EXISTS ( 
                SELECT * FROM sys.databases WHERE name = N'{database_name}'
            )
            BEGIN {create_database} END"""
        create_schema = f"""
            IF NOT EXISTS (
                SELECT * FROM sys.schema WHERE name = N'{schema_name}'
            )
            BEGIN {create_schema} END"""

    if "." in create.schema.prefix:
        # With prefix like "my_db." we create our stages as schemas
        # Attention: we have to rely on a preceding USE statement
        #            for correct prefix database
        return create_schema
    else:
        # With suffix like ".dbo" we create our stages as databases
        return create_database


@compiles(CreateSchema, "ibm_db_sa")
def visit_create_schema_ibm_db_sa(create: CreateSchema, compiler, **kw):
    """For IBM DB2 we need to jump through extra hoops for if_exists=True."""
    _ = kw
    schema = compiler.preparer.format_schema(create.schema.get())
    if create.if_not_exists:
        return (
            "BEGIN\ndeclare continue handler for sqlstate '42710' begin end;\n"
            f"execute immediate 'CREATE SCHEMA {schema}';\nEND"
        )
    else:
        return f"CREATE SCHEMA {schema}"


@compiles(DropSchema)
def visit_drop_schema(drop: DropSchema, compiler, **kw):
    _ = kw
    schema = compiler.preparer.format_schema(drop.schema.get())
    text = ["DROP SCHEMA"]
    if drop.if_exists:
        text.append("IF EXISTS")
    text.append(schema)
    if drop.cascade:
        text.append("CASCADE")
    return " ".join(text)


@compiles(DropSchema, "mssql")
def visit_drop_schema_mssql(drop: DropSchema, compiler, **kw):
    _ = kw
    full_name = drop.schema.get()
    # it was already checked that there is exactly one dot in schema prefix + suffix
    database_name, schema_name = full_name.split(".")
    if "." in drop.schema.prefix:
        # With prefix like "my_db." we create our stages as schemas
        # Attention: we have to rely on a preceding USE statement
        #            for correct prefix database
        text = ["DROP SCHEMA"]
        name = compiler.preparer.format_schema(schema_name)
    else:
        # With suffix like ".dbo" we create our stages as databases
        text = ["DROP DATABASE"]
        name = compiler.preparer.format_schema(database_name)
    if drop.if_exists:
        text.append("IF EXISTS")
    text.append(name)
    return " ".join(text)


# noinspection SqlDialectInspection
@compiles(DropSchema, "ibm_db_sa")
def visit_drop_schema_ibm_db_sa(drop: DropSchema, compiler, **kw):
    """
    Because IBM DB2 doesn't support CASCADE, we must manually drop all tables in
    the schema first.
    """
    statements = []
    if drop.cascade:
        if drop.engine is None:
            raise ValueError(
                "Using DropSchema with cascade=True for ibm_db2 requires passing"
                " the engine kwarg to DropSchema."
            )

        with drop.engine.connect() as conn:
            meta = sa.MetaData()
            meta.reflect(bind=conn, schema=drop.schema.get())

        for table in meta.tables.values():
            statements.append(DropTable(table.name, drop.schema))

    # Compile DROP SCHEMA statement
    schema = compiler.preparer.format_schema(drop.schema.get())
    if drop.if_exists:
        # Add error handler to cache the case that the schema doesn't exist
        statements.append(
            f"""
            BEGIN
                declare continue handler for sqlstate '42704' begin end;
                execute immediate 'DROP SCHEMA {schema} RESTRICT';
            END
            """
        )
    else:
        statements.append(f"DROP SCHEMA {schema} RESTRICT")

    return join_ddl_statements(statements, compiler, **kw)


@compiles(RenameSchema)
def visit_rename_schema(rename: RenameSchema, compiler, **kw):
    _ = kw
    from_ = compiler.preparer.format_schema(rename.from_.get())
    to = compiler.preparer.format_schema(rename.to.get())
    return "ALTER SCHEMA " + from_ + " RENAME TO " + to


@compiles(RenameSchema, "mssql")
def visit_rename_schema_mssql(rename: RenameSchema, compiler, **kw):
    _ = kw
    if rename.from_.prefix != rename.to.prefix:
        raise AttributeError(
            "We currently do not support varying schema prefixes for mssql database"
        )
    from_full_name = rename.from_.get()
    to_full_name = rename.to.get()
    # it was already checked that there is exactly one dot in schema prefix + suffix
    from_database_name, from_schema_name = from_full_name.split(".")
    to_database_name, to_schema_name = to_full_name.split(".")
    if "." in rename.from_.prefix:
        # With prefix like "my_db." we create our stages as schemas
        raise NotImplementedError(
            "There is not SCHEMA rename expression for mssql. "
            "Needs to be addressed on higher level!"
        )
    else:
        # With suffix like ".dbo" we create our stages as databases
        from_name = compiler.preparer.format_schema(from_database_name)
        to_name = compiler.preparer.format_schema(to_database_name)
        return f"ALTER DATABASE {from_name} MODIFY NAME = {to_name}"


@compiles(CreateDatabase)
def visit_create_database(create: CreateDatabase, compiler, **kw):
    _ = kw
    database = compiler.preparer.format_schema(create.database)
    text = ["CREATE DATABASE"]
    if create.if_not_exists:
        text.append("IF NOT EXISTS")
    text.append(database)
    return " ".join(text)


@compiles(DropDatabase)
def visit_drop_database(drop: DropDatabase, compiler, **kw):
    _ = kw
    schema = compiler.preparer.format_schema(drop.database)
    text = ["DROP DATABASE"]
    if drop.if_exists:
        text.append("IF EXISTS")
    text.append(schema)
    if drop.cascade:
        text.append("CASCADE")
    ret = " ".join(text)
    raise NotImplementedError(
        f"Disable for now for safety reasons (not yet needed): {ret}"
    )


def _visit_create_obj_as_select(create, compiler, _type, kw, *, prefix="", suffix=""):
    name = compiler.preparer.quote_identifier(create.name)
    schema = compiler.preparer.format_schema(create.schema.get())
    kw["literal_binds"] = True
    select = compiler.sql_compiler.process(create.query, **kw)
    return f"CREATE {_type} {schema}.{name} AS\n{prefix}{select}{suffix}"


@compiles(CreateTableAsSelect)
def visit_create_table_as_select(create: CreateTableAsSelect, compiler, **kw):
    return _visit_create_obj_as_select(create, compiler, "TABLE", kw)


@compiles(CreateTableAsSelect, "mssql")
def visit_create_table_as_select_mssql(create: CreateTableAsSelect, compiler, **kw):
    name = compiler.preparer.quote_identifier(create.name)
    full_name = create.schema.get()
    # it was already checked that there is exactly one dot in schema prefix + suffix
    database_name, schema_name = full_name.split(".")
    database = compiler.preparer.format_schema(database_name)
    schema = compiler.preparer.format_schema(schema_name)

    kw["literal_binds"] = True
    select = compiler.sql_compiler.process(create.query, **kw)

    return insert_into_in_query(select, database, schema, name)


# noinspection SqlDialectInspection
@compiles(CreateTableAsSelect, "ibm_db_sa")
def visit_create_table_as_select_ibm_db_sa(create: CreateTableAsSelect, compiler, **kw):
    # DB2 stores capitalized table names but sqlalchemy reflects them lowercase
    create = copy.deepcopy(create)
    create.name = ibm_db_sa_fix_name(create.name)

    prepare_statement = _visit_create_obj_as_select(
        create, compiler, "TABLE", kw, prefix="(", suffix=") DEFINITION ONLY"
    )

    name = compiler.preparer.quote_identifier(create.name)
    schema = compiler.preparer.format_schema(create.schema.get())
    kw["literal_binds"] = True
    select = compiler.sql_compiler.process(create.query, **kw)
    create_statement = f"INSERT INTO {schema}.{name}\n{select}"

    return join_ddl_statements([prepare_statement, create_statement], compiler, **kw)


@compiles(CreateViewAsSelect)
def visit_create_view_as_select(create: CreateViewAsSelect, compiler, **kw):
    return _visit_create_obj_as_select(create, compiler, "VIEW", kw)


@compiles(CreateViewAsSelect, "ibm_db_sa")
def visit_create_view_as_select_ibm_db_sa(create: CreateViewAsSelect, compiler, **kw):
    # DB2 stores capitalized table names but sqlalchemy reflects them lowercase
    create = copy.deepcopy(create)
    create.name = ibm_db_sa_fix_name(create.name)
    return _visit_create_obj_as_select(create, compiler, "VIEW", kw)


def insert_into_in_query(select_sql, database, schema, table):
    into = f"INTO {database}.{schema}.{table}"
    into_point = None
    # insert INTO before first FROM, WHERE, GROUP BY, WINDOW, HAVING,
    #                          ORDER BY, UNION, EXCEPT, INTERSECT
    for marker in [
        "FROM",
        "WHERE",
        r"GROUP\s*BY",
        "WINDOW",
        "HAVING",
        r"ORDER\s*BY",
        "UNION",
        "EXCEPT",
        "INTERSECT",
    ]:
        regex = re.compile(marker, re.IGNORECASE)
        for match in regex.finditer(select_sql):
            match_start = match.span()[0]
            prev = select_sql[0:match_start]
            # ignore marker in subqueries in select columns
            if prev.count("(") == prev.count(")"):
                into_point = match_start
                break
        if into_point is not None:
            break
    return (
        select_sql[0:into_point] + into + " " + select_sql[into_point:]
        if into_point is not None
        else select_sql + " " + into
    )


@compiles(CopyTable)
def visit_copy_table(copy_table: CopyTable, compiler, **kw):
    from_name = compiler.preparer.quote_identifier(copy_table.from_name)
    from_schema = compiler.preparer.format_schema(copy_table.from_schema.get())
    query = sa.select("*").select_from(sa.text(f"{from_schema}.{from_name}"))
    create = CreateTableAsSelect(copy_table.to_name, copy_table.to_schema, query)
    return compiler.process(create, **kw)


# noinspection SqlDialectInspection
@compiles(CopyTable, "mssql")
def visit_copy_table_mssql(copy_table: CopyTable, compiler, **kw):
    from_name = compiler.preparer.quote_identifier(copy_table.from_name)
    database, schema = _get_mssql_database_schema(copy_table.from_schema, compiler)
    query = sa.text(f"SELECT * FROM {database}.{schema}.{from_name}")
    create = CreateTableAsSelect(copy_table.to_name, copy_table.to_schema, query)
    return compiler.process(create, **kw)


@compiles(CopyTable, "ibm_db_sa")
def visit_copy_table_ibm_db_sa(copy_table: CopyTable, compiler, **kw):
    copy_table = copy.deepcopy(copy_table)
    copy_table.from_name = ibm_db_sa_fix_name(copy_table.from_name)
    copy_table.to_name = ibm_db_sa_fix_name(copy_table.to_name)
    return visit_copy_table(copy_table, compiler, **kw)


@compiles(DropTable)
def visit_drop_table(drop: DropTable, compiler, **kw):
    return _visit_drop_anything(drop, "TABLE", compiler, **kw)


@compiles(DropTable, "mssql")
def visit_drop_table_mssql(drop: DropTable, compiler, **kw):
    return _visit_drop_anything_mssql(drop, "TABLE", compiler, **kw)


@compiles(DropTable, "ibm_db_sa")
def visit_drop_table_ibm_db_sa(drop: DropTable, compiler, **kw):
    # DB2 stores capitalized table names but sqlalchemy reflects them lowercase
    drop = copy.deepcopy(drop)
    drop.name = ibm_db_sa_fix_name(drop.name)
    return _visit_drop_anything(drop, "TABLE", compiler, **kw)


@compiles(DropView)
def visit_drop_view(drop: DropView, compiler, **kw):
    return _visit_drop_anything(drop, "VIEW", compiler, **kw)


@compiles(DropView, "mssql")
def visit_drop_view_mssql(drop: DropView, compiler, **kw):
    return _visit_drop_anything_mssql(drop, "VIEW", compiler, **kw)


@compiles(DropView, "ibm_db_sa")
def visit_drop_view_ibm_db_sa(drop: DropView, compiler, **kw):
    # DB2 stores capitalized table names but sqlalchemy reflects them lowercase
    drop = copy.deepcopy(drop)
    drop.name = ibm_db_sa_fix_name(drop.name)
    return _visit_drop_anything(drop, "VIEW", compiler, **kw)


@compiles(DropProcedure)
def visit_drop_table(drop: DropProcedure, compiler, **kw):
    return _visit_drop_anything(drop, "PROCEDURE", compiler, **kw)


@compiles(DropProcedure, "mssql")
def visit_drop_table_mssql(drop: DropProcedure, compiler, **kw):
    return _visit_drop_anything_mssql(drop, "PROCEDURE", compiler, **kw)


@compiles(DropFunction)
def visit_drop_table(drop: DropFunction, compiler, **kw):
    return _visit_drop_anything(drop, "FUNCTION", compiler, **kw)


@compiles(DropFunction, "mssql")
def visit_drop_table_mssql(drop: DropProcedure, compiler, **kw):
    return _visit_drop_anything_mssql(drop, "FUNCTION", compiler, **kw)


def _visit_drop_anything(
    drop: DropTable | DropView | DropProcedure | DropFunction,
    _type,
    compiler,
    dont_quote_table=False,
    **kw,
):
    _ = kw
    if dont_quote_table:
        table = drop.name
    else:
        table = compiler.preparer.quote_identifier(drop.name)
    schema = compiler.preparer.format_schema(drop.schema.get())
    text = [f"DROP {_type}"]
    if drop.if_exists:
        text.append("IF EXISTS")
    text.append(f"{schema}.{table}")
    return " ".join(text)


def _visit_drop_anything_mssql(
    drop: DropTable | DropView | DropProcedure | DropFunction, _type, compiler, **kw
):
    _ = kw
    table = compiler.preparer.quote_identifier(drop.name)
    database, schema = _get_mssql_database_schema(drop.schema, compiler)
    text = [f"DROP {_type}"]
    if drop.if_exists:
        text.append("IF EXISTS")
    if isinstance(drop, (DropView, DropProcedure, DropFunction)):
        # attention: this statement must be prefixed with a 'USE <database>' statement
        text.append(f"{schema}.{table}")
    else:
        text.append(f"{database}.{schema}.{table}")
    return " ".join(text)


# noinspection SqlDialectInspection
@compiles(AddPrimaryKey)
def visit_add_primary_key(add_primary_key: AddPrimaryKey, compiler, **kw):
    _ = kw
    table = compiler.preparer.quote_identifier(add_primary_key.table_name)
    schema = compiler.preparer.format_schema(add_primary_key.schema.get())
    pk_name = compiler.preparer.quote_identifier(
        add_primary_key.name
        if add_primary_key.name is not None
        else "pk_" + "_".join(add_primary_key.key)
    )
    cols = ",".join(
        [compiler.preparer.quote_identifier(col) for col in add_primary_key.key]
    )
    return f"ALTER TABLE {schema}.{table} ADD CONSTRAINT {pk_name} PRIMARY KEY ({cols})"


# noinspection SqlDialectInspection
@compiles(AddPrimaryKey, "mssql")
def visit_add_primary_key(add_primary_key: AddPrimaryKey, compiler, **kw):
    _ = kw
    database, schema = _get_mssql_database_schema(add_primary_key.schema, compiler)

    table = compiler.preparer.quote_identifier(add_primary_key.table_name)
    pk_name = compiler.preparer.quote_identifier(
        add_primary_key.name
        if add_primary_key.name is not None
        else "pk_" + "_".join(add_primary_key.key)
    )
    cols = ",".join(
        [compiler.preparer.quote_identifier(col) for col in add_primary_key.key]
    )
    return (
        f"ALTER TABLE {database}.{schema}.{table} ADD CONSTRAINT {pk_name} PRIMARY KEY"
        f" ({cols})"
    )


# noinspection SqlDialectInspection
@compiles(AddPrimaryKey, "ibm_db_sa")
def visit_add_primary_key(add_primary_key: AddPrimaryKey, compiler, **kw):
    _ = kw
    # DB2 stores capitalized table names but sqlalchemy reflects them lowercase
    add_primary_key = copy.deepcopy(add_primary_key)
    add_primary_key.table_name = ibm_db_sa_fix_name(add_primary_key.table_name)

    table = compiler.preparer.quote_identifier(add_primary_key.table_name)
    schema = compiler.preparer.format_schema(add_primary_key.schema.get())
    pk_name = compiler.preparer.quote_identifier(
        add_primary_key.name
        if add_primary_key.name is not None
        else "pk_" + "_".join(add_primary_key.key)
    )
    cols = ",".join(
        [
            compiler.preparer.quote_identifier(ibm_db_sa_fix_name(col))
            for col in add_primary_key.key
        ]
    )
    return f"ALTER TABLE {schema}.{table} ADD CONSTRAINT {pk_name} PRIMARY KEY ({cols})"


# noinspection SqlDialectInspection
@compiles(AddIndex)
def visit_add_index(add_index: AddIndex, compiler, **kw):
    _ = kw
    table = compiler.preparer.quote_identifier(add_index.table_name)
    schema = compiler.preparer.format_schema(add_index.schema.get())
    index_name = compiler.preparer.quote_identifier(
        add_index.name
        if add_index.name is not None
        else "idx_" + "_".join(add_index.index)
    )
    cols = ",".join(
        [compiler.preparer.quote_identifier(col) for col in add_index.index]
    )
    return f"CREATE INDEX {index_name} ON {schema}.{table} ({cols})"


# noinspection SqlDialectInspection
@compiles(AddIndex, "mssql")
def visit_add_index(add_index: AddIndex, compiler, **kw):
    _ = kw
    database, schema = _get_mssql_database_schema(add_index.schema, compiler)

    table = compiler.preparer.quote_identifier(add_index.table_name)
    index_name = compiler.preparer.quote_identifier(
        add_index.name
        if add_index.name is not None
        else "idx_" + "_".join(add_index.index)
    )
    cols = ",".join(
        [compiler.preparer.quote_identifier(col) for col in add_index.index]
    )
    return f"CREATE INDEX {index_name} ON {database}.{schema}.{table} ({cols})"


# noinspection SqlDialectInspection
@compiles(AddIndex, "ibm_db_sa")
def visit_add_index(add_index: AddIndex, compiler, **kw):
    _ = kw
    # DB2 stores capitalized table names but sqlalchemy reflects them lowercase
    add_index = copy.deepcopy(add_index)
    add_index.table_name = ibm_db_sa_fix_name(add_index.table_name)

    table = compiler.preparer.quote_identifier(add_index.table_name)
    schema = compiler.preparer.format_schema(add_index.schema.get())
    index_name = compiler.preparer.quote_identifier(
        add_index.name
        if add_index.name is not None
        else "idx_" + "_".join(add_index.index)
    )
    cols = ",".join(
        [
            compiler.preparer.quote_identifier(ibm_db_sa_fix_name(col))
            for col in add_index.index
        ]
    )
    return f"CREATE INDEX {schema}.{index_name} ON {schema}.{table} ({cols})"


@compiles(ChangeColumnTypes)
def visit_change_column_types(change: ChangeColumnTypes, compiler, **kw):
    _ = kw
    table = compiler.preparer.quote_identifier(change.table_name)
    schema = compiler.preparer.format_schema(change.schema.get())
    alter_columns = ",".join(
        [
            f"ALTER COLUMN {compiler.preparer.quote_identifier(col)} SET DATA TYPE"
            f" {compiler.type_compiler.process(_type)}"
            for col, _type, nullable in zip(
                change.column_names, change.column_types, change.nullable
            )
        ]
        + [
            "ALTER COLUMN"
            f" {compiler.preparer.quote_identifier(col)}"
            f" {'SET' if not nullable else 'DROP'} NOT NULL"
            for col, nullable in zip(change.column_names, change.nullable)
            if nullable is not None
        ]
    )
    return f"ALTER TABLE {schema}.{table} {alter_columns}"


# noinspection SqlDialectInspection
@compiles(ChangeColumnTypes, "mssql")
def visit_change_column_types(change: ChangeColumnTypes, compiler, **kw):
    _ = kw
    database, schema = _get_mssql_database_schema(change.schema, compiler)

    table = compiler.preparer.quote_identifier(change.table_name)

    def modify_type(_type):
        if change.cap_varchar_max is not None:
            _type = copy.copy(_type)
            if isinstance(_type, sa.String) and (
                _type.length is None or _type.length > change.cap_varchar_max
            ):
                # impose some limit to allow use in primary key / index
                _type.length = change.cap_varchar_max
        return _type

    statements = [
        f"ALTER TABLE {database}.{schema}.{table} ALTER COLUMN"
        f" {compiler.preparer.quote_identifier(col)} "
        f"{compiler.type_compiler.process(modify_type(_type))}"
        f"{'' if nullable is None else ' NULL' if nullable else ' NOT NULL'}"
        for col, _type, nullable in zip(
            change.column_names, change.column_types, change.nullable
        )
    ]
    return join_ddl_statements(statements, compiler, **kw)


# noinspection SqlDialectInspection
@compiles(ChangeColumnTypes, "ibm_db_sa")
def visit_change_column_types(change: ChangeColumnTypes, compiler, **kw):
    _ = kw
    # DB2 stores capitalized table names but sqlalchemy reflects them lowercase
    change = copy.deepcopy(change)
    change.table_name = ibm_db_sa_fix_name(change.table_name)

    def modify_type(_type):
        if change.cap_varchar_max is not None:
            _type = copy.copy(_type)
            if isinstance(_type, sa.String) and (
                _type.length is None or _type.length > change.cap_varchar_max
            ):
                # impose some limit to allow use in primary key / index
                _type.length = change.cap_varchar_max
        return _type

    table = compiler.preparer.quote_identifier(change.table_name)
    schema = compiler.preparer.format_schema(change.schema.get())
    statements = [
        f"ALTER TABLE {schema}.{table} ALTER COLUMN"
        f" {compiler.preparer.quote_identifier(ibm_db_sa_fix_name(col))} SET DATA TYPE"
        f" {compiler.type_compiler.process(modify_type(_type))}"
        for col, _type, nullable in zip(
            change.column_names, change.column_types, change.nullable
        )
    ] + [
        f"ALTER TABLE {schema}.{table} ALTER COLUMN"
        f" {compiler.preparer.quote_identifier(ibm_db_sa_fix_name(col))}"
        f" {'SET' if not nullable else 'DROP'} NOT NULL"
        for col, nullable in zip(change.column_names, change.nullable)
        if nullable is not None
    ]
    statements.append(f"call sysproc.admin_cmd('REORG TABLE {schema}.{table}')")
    return join_ddl_statements(statements, compiler, **kw)


# noinspection SqlDialectInspection
@compiles(ChangeColumnNullable)
def visit_change_column_types(change: ChangeColumnNullable, compiler, **kw):
    _ = kw
    table = compiler.preparer.quote_identifier(change.table_name)
    schema = compiler.preparer.format_schema(change.schema.get())
    alter_columns = ",".join(
        [
            "ALTER COLUMN"
            f" {compiler.preparer.quote_identifier(col)}"
            f" {'SET' if not nullable else 'DROP'} NOT NULL"
            for col, nullable in zip(change.column_names, change.nullable)
        ]
    )
    return f"ALTER TABLE {schema}.{table} {alter_columns}"


# noinspection SqlDialectInspection
@compiles(ChangeColumnNullable, "ibm_db_sa")
def visit_change_column_types(change: ChangeColumnNullable, compiler, **kw):
    _ = kw
    # DB2 stores capitalized table names but sqlalchemy reflects them lowercase
    change = copy.deepcopy(change)
    change.table_name = ibm_db_sa_fix_name(change.table_name)

    table = compiler.preparer.quote_identifier(change.table_name)
    schema = compiler.preparer.format_schema(change.schema.get())
    statements = [
        f"ALTER TABLE {schema}.{table} ALTER COLUMN"
        f" {compiler.preparer.quote_identifier(ibm_db_sa_fix_name(col))}"
        f" {'SET' if not nullable else 'DROP'} NOT NULL"
        for col, nullable in zip(change.column_names, change.nullable)
    ]
    statements.append(f"call sysproc.admin_cmd('REORG TABLE {schema}.{table}')")
    return join_ddl_statements(statements, compiler, **kw)


def ibm_db_sa_fix_name(name):
    # DB2 seems to create tables uppercase if all lowercase given
    return name.upper() if name.islower() else name


STATEMENT_SEPERATOR = "; -- PYDIVERSE-PIPEDAG-SPLIT\n"


def join_ddl_statements(statements, compiler, **kw):
    """Mechanism to combine multiple DDL statements into one."""
    statement_strings = []
    kw["literal_binds"] = True
    for statement in statements:
        if isinstance(statement, str):
            statement_strings.append(statement)
        else:
            statement_strings.append(compiler.process(statement, **kw))
    return STATEMENT_SEPERATOR.join(statement_strings)


def split_ddl_statement(statement: str):
    """Split previously combined DDL statements apart"""
    return statement.split(STATEMENT_SEPERATOR)


def _get_mssql_database_schema(schema: Schema, compiler):
    full_name = schema.get()
    # it was already checked that there is exactly one dot in schema prefix + suffix
    database_name, schema_name = full_name.split(".")
    database = compiler.preparer.format_schema(database_name)
    schema = compiler.preparer.format_schema(schema_name)
    return database, schema



def normalize_name(name: str) -> str:
    """Normalizes an identifier

    All names in PipeDAG are case-insensitive and can't contain any
    slashes. This helper function does exactly this conversion.
    """
    if name is not None:
        return name.casefold().strip().replace("/", "_")


def safe_name(name: str) -> str:
    """Converts an identifier to one that is lowercase, ascii only

    Some backends might only support a limited set of characters for
    identifiers. This generic functions provides a mechanism for making
    a name safe (at least in most bases) by encoding non ascii characters
    using punycode.

    :param name: The identifier / name to make safe
    :return: The safe name
    """
    name = normalize_name(name)
    name = name.encode("punycode").decode("ascii")
    return name



from __future__ import annotations

import copy
import getpass
import itertools
import logging
import os
import re
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterable

import structlog
import yaml
from box import Box

from pydiverse.pipedag.util.deep_merge import deep_merge
from pydiverse.pipedag.util.import_ import import_object

if TYPE_CHECKING:
    from pydiverse.pipedag.context import ConfigContext


# noinspection PyPep8Naming
class cached_class_property:
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, cls):
        if not hasattr(self, "cache"):
            self.cache = self.func(cls)
        return self.cache


class PipedagConfig:
    default: PipedagConfig

    def __init__(self, path: str):
        self.path = path
        with open(path) as f:
            self.raw_config = yaml.safe_load(f)

        self.config_dict = self.__parse_config(self.raw_config)

    @cached_class_property
    def default(cls):
        config_path = find_config()
        return PipedagConfig(config_path)

    @property
    def name(self) -> str:
        return self.config_dict.get("name")

    def __parse_config(self, raw_config: dict[str:Any]):
        config = copy.deepcopy(raw_config)

        # Expand all references
        for instance_dict in _get(config, "instances", default={}).values():
            self.__expand_references(instance_dict)

        for flow_dict in _get(config, "flows", default={}).values():
            self.__expand_references(flow_dict)
            for instance_dict in _get(flow_dict, "instances", default={}).values():
                self.__expand_references(instance_dict)

        return config

    def __expand_references(self, config):
        references = [
            (None, "technical_setup", "technical_setups"),
            ("table_store", "table_store_connection", "table_store_connections"),
            ("blob_store", "blob_store_connection", "blob_store_connections"),
        ]

        for expand_path, ref_name_path, ref_src_path in references:
            ref_name = _get(config, expand_path, ref_name_path, default=None)
            if ref_name is None:
                continue

            base_dict = _get(config, expand_path)
            merged_dict = deep_merge(
                base_dict, copy.deepcopy(_get(self.raw_config, ref_src_path, ref_name))
            )
            base_dict.update(merged_dict)
            _pop(base_dict, ref_name_path)

        return config

    def get(
        self,
        instance: str | None = None,
        flow: str | None = None,
        per_user: bool = False,
    ) -> ConfigContext:
        # TODO: Check that this function only gets called in the main interpreter.
        #       Otherwise certain environment variables might get expanded incorrectly.
        from pydiverse.pipedag.context import ConfigContext

        config = self.__get_merged_config_dict(
            instance=instance,
            flow=flow,
            default={
                "fail_fast": False,
                "network_interface": "127.0.0.1",
                "per_user_template": "{id}_{username}",
                "strict_result_get_locking": True,
                "stage_commit_technique": "SCHEMA_SWAP",
                "auto_table": [],
                "auto_blob": [],
                "attrs": {},
            },
        )

        # check enums
        # Alternative: could be a feature of __get_merged_config_dict
        # in case default value is set to Enum
        from pydiverse.pipedag.context.context import StageCommitTechnique

        config["stage_commit_technique"] = (
            config["stage_commit_technique"].strip().upper()
        )
        if not hasattr(StageCommitTechnique, config["stage_commit_technique"]):
            raise ValueError(
                "Found unknown setting stage_commit_technique:"
                f" '{config['stage_commit_technique']}'; Expected one of:"
                f" {', '.join([v.name for v in StageCommitTechnique])}"
            )
        stage_commit_technique = getattr(
            StageCommitTechnique, config["stage_commit_technique"]
        )

        # TODO: Delegate selecting where variables can be expanded to the
        #  corresponding classes.
        #    eg. SQLTableStore._expand_env_vars = ["url", "url_attrs_file"]
        #    eg. SQLTableStore._expand_vars = ["url", "schema_prefix", "schema_suffix"]

        # First expand all environment variables
        self.__expand_environment_variables(inout_config=config)

        # Intermediate variable processing
        if per_user:
            template = config["per_user_template"]
            config["instance_id"] = expand_variables(
                template,
                {
                    "username": getpass.getuser(),
                    "id": config["instance_id"],
                },
            )

        # Handle url_attrs_file
        url_attrs_file = _pop(config, "table_store", "url_attrs_file", default=None)
        if url_attrs_file is not None:
            with open(url_attrs_file, encoding="utf-8") as fh:
                url_attrs = yaml.safe_load(fh)

            url = _get(config, "table_store", "url")
            url = expand_variables(url, url_attrs, skip_missing=True)
            _set(config, url, "table_store", "url")

        # Finally, expand all normal variables
        self.__expand_variables(inout_config=config)

        # Construct final ConfigContext
        config_context = ConfigContext(
            config_dict=config,
            pipedag_name=self.name,
            flow_name=flow,
            strict_result_get_locking=config["strict_result_get_locking"],
            instance_name=instance,
            instance_id=config["instance_id"],
            stage_commit_technique=stage_commit_technique,
            fail_fast=config["fail_fast"],
            network_interface=config["network_interface"],
            attrs=Box(config["attrs"], frozen_box=True),
        )

        # Make sure @cached_property store is set up and loaded
        # and throw config errors early.
        with config_context:
            _ = config_context.store
            _ = config_context.auto_table
            _ = config_context.auto_blob

            config_context.create_orchestration_engine().dispose()
            config_context.create_lock_manager().dispose()

        return config_context

    def __get_merged_config_dict(self, instance, flow, default=None):
        search_paths = [
            ("instances", "__any__"),
            ("instances", instance),
            ("flows", "__any__", "instances", "__any__"),
            ("flows", "__any__", "instances", instance),
            ("flows", flow, "instances", "__any__"),
            ("flows", flow, "instances", instance),
            ("flows", flow),
        ]

        search_paths = [path for path in search_paths if None not in path]
        dicts = [_get(self.config_dict, path, default=None) for path in search_paths]

        # Check for strict instance lookup
        # If instance is specified, make sure that a corresponding section can be found
        strict_instance_lookup = self.config_dict.get("strict_instance_lookup", True)
        if strict_instance_lookup and instance is not None:
            found_instance = False
            for path, d in zip(search_paths, dicts):
                found_instance |= (
                    "instances" in path and instance in path and d is not None
                )

            if not found_instance:
                raise AttributeError(
                    "Strict instance lookup failed: Couldn't find instance"
                    f" '{instance}' in pipedag config."
                )

        # Merge
        merged = default or {}
        for d in dicts:
            if d is not None:
                merged = deep_merge(merged, d)

        return merged

    @staticmethod
    def __expand_environment_variables(*, inout_config):
        locations = [
            ("table_store", "url"),
            ("table_store", "url_attrs_file"),
        ]

        for location in locations:
            value: str = _get(inout_config, location, default=None)
            if value is None:
                continue

            value = expand_environment_variables(value)
            _set(inout_config, value, location)

    def __expand_variables(self, *, inout_config):
        locations = [
            ("table_store", "url"),
            ("table_store", "schema_prefix"),
            ("table_store", "schema_suffix"),
        ]

        # TODO: Decide on a list of available variables
        variables = {
            "username": getpass.getuser(),
            "instance_id": inout_config.get("instance_id"),
            "name": self.name,
        }

        for key, val in variables.items():
            if val is None:
                variables.pop(key)

        for location in locations:
            value: str = _get(inout_config, location, default=None)
            if value is None:
                continue

            value = expand_variables(value, variables)
            _set(inout_config, value, location)


def find_config(
    name: str = "pipedag",
    search_paths: Iterable[str | Path] = None,
) -> str:
    """Searches for a pipedag config file

    The following paths get checked first.

    - The path specified in the "PIPEDAG_CONFIG" environment variable

    Else it searches in the following locations:

    - Current working directory
    - All parent directories
    - The user folder

    :param name: The name of the config file
    :param search_paths: The directories in which to search for the config file
    :return: The path of the file.
    :raises FileNotFoundError: if no config file could be found.
    """

    extensions = [".yaml", ".yml"]

    if search_paths is None:
        # Check PIPEDAG_CONFIG path
        if path := os.environ.get("PIPEDAG_CONFIG", None):
            path = Path(path).resolve().expanduser()
            if path.is_file():
                return str(path)

            for extension in extensions:
                path = path / (name + extension)
                if path.is_file():
                    return str(path)

        # Else, search in these default directories
        search_paths = [
            Path.cwd(),
            *Path.cwd().resolve().parents,
            Path("~").expanduser(),
        ]
    else:
        search_paths = [Path(path) for path in search_paths]

    file_names = [name + extension for extension in extensions]
    for path, file_name in itertools.product(search_paths, file_names):
        config_path = (path / file_name).resolve()
        if config_path.is_file():
            return str(config_path)

    raise FileNotFoundError("No config file found")


def expand_environment_variables(string: str) -> str:
    """
    Expands all occurrences of the form `{$ENV_VAR}` with the environment variable
    named `ENV_VAR`.
    """

    def env_var_sub(match: re.Match):
        name = match.group()[2:-1]
        if name not in os.environ:
            raise AttributeError(
                f"Could not find environment variable '{name}' "
                f"referenced in '{string}'."
            )
        return os.environ[name]

    return re.sub(r"\{\$[a-zA-Z_]+[a-zA-Z0-9_]*\}", env_var_sub, string)


def expand_variables(
    string: str, variables: dict[str:str], skip_missing: bool = False
) -> str:
    """
    Expands all occurrences of the form {var_name} with the variable
    named `var_name`.
    """

    def var_sub(match: re.Match):
        name = match.group()[1:-1]
        if name not in variables:
            if skip_missing:
                return match.group()
            raise AttributeError(
                f"Could not find variable '{name}' referenced in '{string}'."
            )
        return str(variables[name])

    return re.sub(r"\{[a-zA-Z_]+[a-zA-Z0-9_]*\}", var_sub, string)


def load_object(config_dict: dict):
    """Instantiates an instance of an object given

    The import path (module.Class) should be specified as the "class" value
    of the dict. The rest of the dict get used as the instance config.

    If the class defines a `_init_conf_` function, it gets called using the
    config values, otherwise they just get passed to the class initializer.

    >>> # module.Class(argument="value")
    >>> load_object({
    >>>     "class": "module.Class",
    >>>     "argument": "value",
    >>> })
    """

    config_dict = config_dict.copy()
    cls = import_object(config_dict.pop("class"))

    try:
        init_conf = cls._init_conf_
        return init_conf(config_dict)
    except AttributeError:
        return cls(**config_dict)


# Nested Dictionary Utilities


def _flatten(items):
    for x in items:
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            yield from _flatten(x)
        else:
            yield x


_nil = object()


def _get(d, *path, default=_nil):
    path = [x for x in _flatten(path) if x is not None]
    try:
        for p in path:
            if p is not None:
                d = d[p]
    except (IndexError, KeyError) as e:
        if default is _nil:
            raise KeyError(f"Invalid path: {' > '.join(path)}") from e
        return default

    return d


def _set(d, value, *path):
    path = [x for x in _flatten(path) if x is not None]
    for p in path[:-1]:
        d = d[p]
    d[path[-1]] = value


def _pop(d, *path, default=_nil) -> Any:
    try:
        path = [x for x in _flatten(path) if x is not None]
        for p in path[:-1]:
            d = d[p]
        return d.pop(path[-1])
    except KeyError as e:
        if default == _nil:
            raise e
        return default


def setup_structlog(_log_level=logging.INFO, _log_stream=sys.stderr):
    logging.basicConfig(
        stream=_log_stream,
        format="%(asctime)s [%(levelname)s] %(message)s",
        level=_log_level,
    )
    structlog.configure(
        processors=[
            structlog.contextvars.merge_contextvars,
            structlog.processors.add_log_level,
            structlog.processors.StackInfoRenderer(),
            structlog.dev.set_exc_info,
            structlog.processors.TimeStamper(),
            structlog.dev.ConsoleRenderer(),
        ],
        wrapper_class=structlog.make_filtering_bound_logger(_log_level),
        context_class=dict,
        logger_factory=structlog.PrintLoggerFactory(_log_stream),
        cache_logger_on_first_use=True,
    )



from __future__ import annotations

from typing import TYPE_CHECKING

import structlog

from pydiverse.pipedag.context import ConfigContext, DAGContext
from pydiverse.pipedag.core.task import Task
from pydiverse.pipedag.errors import StageError
from pydiverse.pipedag.util import normalize_name

if TYPE_CHECKING:
    from pydiverse.pipedag.core.flow import Flow


class Stage:
    """The Stage class is used to group a collection of related tasks

    The main purpose of a Stage is to allow for a transactionality mechanism.
    Only if all tasks inside a stage finish successfully does the stage
    get committed.

    All task that get defined inside the stage's context will automatically
    get added to the stage.

    An example of how to use a Stage:
    ::

        @materialize()
        def my_materializing_task():
            return Table(...)

        with Flow("my_flow") as flow:
            with Stage("stage_1") as stage_1:
                task_1 = my_materializing_task()
                task_2 = another_materializing_task(task_1)

            with Stage("stage_2") as stage_2:
                task_3 = yet_another_task(task_3)
                ...

        flow.run()

    To ensure that all tasks get executed in the correct order, each
    MaterializingTask must be an upstream dependency of its stage's
    CommitStageTask (this is done by calling the `add_task` method) and
    for each of its upstream dependencies, the associated StageCommitTask
    must be added as an upstream dependency.
    This ensures that the stage commit only happens after all tasks have
    finished writing to the transaction stage, and a task never gets executed
    before any of its upstream stage dependencies have been committed.

    :param name: The name of the stage. Two stages with the same name may
        not be used inside the same flow.
    """

    def __init__(self, name: str):
        self._name = normalize_name(name)
        self._transaction_name = f"{self._name}__tmp"

        self.tasks: list[Task] = []
        self.commit_task: CommitStageTask = None  # type: ignore
        self.outer_stage: Stage | None = None

        self.logger = structlog.get_logger(stage=self)
        self.id: int = None  # type: ignore

        self._did_enter = False

    @property
    def name(self) -> str:
        return self._name

    @property
    def transaction_name(self) -> str:
        return self._transaction_name

    def set_transaction_name(self, new_transaction_name):
        # used by stage_commit_technique=READ_VIEWS to change odd/even
        # transaction schemas
        self._transaction_name = new_transaction_name

    @property
    def current_name(self) -> str:
        """The name of the stage where the data currently lives

        Before a task has been committed this is the transaction name,
        after the commit it is the base name.
        """

        if self.did_commit:
            return self.name
        else:
            return self.transaction_name

    @property
    def did_commit(self) -> bool:
        from pydiverse.pipedag.context.run_context import RunContext, StageState

        return RunContext.get().get_stage_state(self) == StageState.COMMITTED

    def __repr__(self):
        return f"<Stage: {self.name}>"

    def __enter__(self):
        if self._did_enter:
            raise StageError(
                f"Stage '{self.name}' has already been entered."
                " Can't reuse the same stage twice."
            )
        self._did_enter = True

        # Capture information from surrounding Flow or Stage block
        # and link this stage with it
        try:
            outer_ctx = DAGContext.get()
        except LookupError as e:
            raise StageError("Stage can't be defined outside of a flow") from e

        outer_ctx.flow.add_stage(self)
        if outer_ctx.stage is not None:
            self.outer_stage = outer_ctx.stage

        # Initialize new context (both Flow and Stage use DAGContext to transport
        # information to @materialize annotations within the flow and to support
        # nesting of stages)
        self._ctx = DAGContext(
            flow=outer_ctx.flow,
            stage=self,
        )
        self._ctx.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.commit_task = CommitStageTask(self, self._ctx.flow)
        self._ctx.flow.add_task(self.commit_task)
        self._ctx.__exit__()
        del self._ctx

    def all_tasks(self):
        yield from self.tasks
        yield self.commit_task

    def is_inner(self, other: Stage):
        outer = self.outer_stage
        while outer is not None:
            if outer == other:
                return True
            outer = outer.outer_stage
        return False

    def __getstate__(self):
        state = self.__dict__.copy()
        state.pop("tasks", None)
        state.pop("commit_task", None)
        state.pop("outer_stage", None)
        state.pop("logger", None)
        return state


class CommitStageTask(Task):
    def __init__(self, stage: Stage, flow: Flow):
        super().__init__(
            fn=self.fn,
            name=f"Commit '{stage.name}'",
        )

        self.stage = stage
        self.flow = flow
        self.upstream_stages = {stage}
        self.input_tasks = {}

        self._bound_args = self._signature.bind()
        self._visualize_hidden = True

    def fn(self):
        self.logger.info(f"Committing stage '{self.stage.name}'")
        ConfigContext.get().store.commit_stage(self.stage)



from __future__ import annotations

import warnings

import pandas as pd

from pydiverse.pipedag import Stage, Table
from pydiverse.pipedag.backend.table.base import BaseTableStore, TableHook
from pydiverse.pipedag.errors import CacheError, StageError
from pydiverse.pipedag.materialize.core import MaterializingTask
from pydiverse.pipedag.materialize.metadata import LazyTableMetadata, TaskMetadata


class DictTableStore(BaseTableStore):
    """
    A very basic table store that stores objects in a dictionary.
    Should only ever be used for testing.
    """

    def __init__(self):
        super().__init__()
        self.store = dict()

        self.metadata = dict()
        self.t_metadata = dict()

        self.lazy_table_metadata = dict()
        self.t_lazy_table_metadata = dict()

    def init_stage(self, stage: Stage):
        self.store.setdefault(stage.name, {})
        self.store[stage.transaction_name] = {}

        self.metadata.setdefault(stage, {})
        self.t_metadata[stage] = {}

        self.lazy_table_metadata.setdefault(stage, {})
        self.t_lazy_table_metadata[stage] = {}

    def commit_stage(self, stage: Stage):
        self.store[stage.name] = self.store[stage.transaction_name]
        del self.store[stage.transaction_name]
        self.metadata[stage] = self.t_metadata[stage]
        self.lazy_table_metadata[stage] = self.t_lazy_table_metadata[stage]

    def copy_table_to_transaction(self, table: Table):
        stage = table.stage
        if stage.did_commit:
            raise StageError(
                f"Can't copy table '{table.name}' to transaction."
                f" Stage '{stage.name}' has already been committed."
            )

        try:
            t = self.store[stage.name][table.name]
            self.store[stage.transaction_name][table.name] = t
        except KeyError:
            raise CacheError(
                f"No table with name '{table.name}' found in '{stage.name}' stage"
            ) from None

    def copy_lazy_table_to_transaction(self, metadata: LazyTableMetadata, table: Table):
        stage = table.stage
        if stage.did_commit:
            raise StageError(
                f"Can't copy table '{metadata.name}' to transaction."
                f" Stage '{stage.name}' has already been committed."
            )

        try:
            t = self.store[stage.name][metadata.name]
            self.store[stage.transaction_name][metadata.name] = t
        except KeyError:
            raise CacheError(
                f"No table with name '{metadata.name}' found in '{stage.name}' stage"
            ) from None

    def delete_table_from_transaction(self, table: Table):
        try:
            self.store[table.stage.transaction_name].pop(table.name)
        except KeyError:
            return

    def store_task_metadata(self, metadata: TaskMetadata, stage: Stage):
        cache_key = metadata.input_hash + str(metadata.cache_fn_hash)
        self.t_metadata[stage][cache_key] = metadata

    def retrieve_task_metadata(
        self, task: MaterializingTask, input_hash: str, cache_fn_hash: str
    ) -> TaskMetadata:
        cache_key = input_hash + str(cache_fn_hash)
        try:
            return self.metadata[task.stage][cache_key]
        except KeyError:
            raise CacheError(
                "There is no metadata for task "
                f"'{task.name}' with cache key '{task.input_hash}', yet"
            ) from None

    def store_lazy_table_metadata(self, metadata: LazyTableMetadata):
        cache_key = metadata.query_hash + metadata.task_hash
        self.t_lazy_table_metadata[metadata.stage][cache_key] = metadata

    def retrieve_lazy_table_metadata(
        self, query_hash: str, task_hash: str, stage: Stage
    ) -> LazyTableMetadata:
        try:
            cache_key = query_hash + task_hash
            return self.lazy_table_metadata[stage.name][cache_key]
        except (TypeError, KeyError):
            raise CacheError("Couldn't find metadata for lazy table") from None

    def list_tables(self, stage, *, include_everything=False):
        """
        List all tables that were generated in a stage.

        It may also include other objects database objects like views, stored
        procedures, functions, etc. which makes the name `list_tables` too specific.
        But the predominant idea is that tasks produce tables in stages and thus the
        storyline of callers is much nicer to read. In the end we might need everything
        to recover the full cache output which was produced by a RawSQL statement
        (we want to be compatible with legacy sql code as a starting point).

        :param stage: the stage
        :param include_everything: If True, we might include stored procedures,
            functions and other database objects that have a schema associated name.
        :return: list of tables [and other objects]
        """
        return self.store[stage.transaction_name].keys()


@DictTableStore.register_table(pd)
class PandasTableHook(TableHook[DictTableStore]):
    @classmethod
    def can_materialize(cls, type_) -> bool:
        return issubclass(type_, pd.DataFrame)

    @classmethod
    def can_retrieve(cls, type_) -> bool:
        return type_ == pd.DataFrame

    @classmethod
    def materialize(cls, store, table: Table[pd.DataFrame], stage_name):
        if table.name is not None:
            table.obj.attrs["name"] = table.name
        store.store[stage_name][table.name] = table.obj

    @classmethod
    def retrieve(cls, store, table, stage_name, as_type):
        return store.store[stage_name][table.name].copy()

    @classmethod
    def auto_table(cls, obj: pd.DataFrame):
        if name := obj.attrs.get("name"):
            return Table(obj, name)
        return super().auto_table(obj)


try:
    import pydiverse.transform as pdt
except ImportError as e:
    warnings.warn(str(e), ImportWarning)
    pdt = None


@DictTableStore.register_table(pdt)
class PydiverseTransformTableHook(TableHook[DictTableStore]):
    @classmethod
    def can_materialize(cls, type_) -> bool:
        return issubclass(type_, pdt.Table)

    @classmethod
    def can_retrieve(cls, type_) -> bool:
        from pydiverse.transform.eager import PandasTableImpl

        return issubclass(type_, PandasTableImpl)

    @classmethod
    def materialize(cls, store, table: Table[pdt.Table], stage_name):
        from pydiverse.transform.core.verbs import collect

        table.obj = table.obj >> collect()
        # noinspection PyTypeChecker
        return PandasTableHook.materialize(store, table, stage_name)

    @classmethod
    def retrieve(cls, store, table, stage_name, as_type):
        from pydiverse.transform.eager import PandasTableImpl

        df = PandasTableHook.retrieve(store, table, stage_name, pd.DataFrame)
        return pdt.Table(PandasTableImpl(table.name, df))

    @classmethod
    def auto_table(cls, obj: pdt.Table):
        # noinspection PyProtectedMember
        return Table(obj, obj._impl.name)



from pydiverse.pipedag.errors import DisposedError


class Disposable:
    def __getattribute__(self, name):
        try:
            object.__getattribute__(self, "_Disposable__disposed")
            obj_type = object.__getattribute__(self, "__class__")
            raise DisposedError(f"Object of type {obj_type} has already been disposed.")
        except AttributeError:
            return object.__getattribute__(self, name)

    def __setattr__(self, key, value):
        try:
            object.__getattribute__(self, "_Disposable__disposed")
            obj_type = object.__getattribute__(self, "__class__")
            raise DisposedError(f"Object of type {obj_type} has already been disposed.")
        except AttributeError:
            return object.__setattr__(self, key, value)

    def dispose(self):
        object.__setattr__(self, "_Disposable__disposed", True)



from __future__ import annotations

import builtins
import importlib
from typing import Any


def requires(requirements: Any | list, exception: BaseException | type[BaseException]):
    """Class decorator for handling optional imports.

    If any of the requirements are falsy, this decorator prevents the class
    from being instantiated and any class attributes from being accessed,
    and raises the provided exception instead.
    """

    if not isinstance(requirements, (list, tuple)):
        requirements = (requirements,)

    def decorator(cls):
        if all(requirements):
            return cls

        # Modify class to raise exception
        class RaiserMeta(type):
            def __getattr__(self, x):
                raise exception

        def raiser(*args, **kwargs):
            raise exception

        __name = str(cls.__name__)
        __bases = ()
        __dict = {
            "__metaclass__": RaiserMeta,
            "__wrapped__": cls,
            "__new__": raiser,
        }

        return RaiserMeta(__name, __bases, __dict)

    return decorator


def import_object(import_path: str):
    """Loads a class given an import path

    >>> # An import statement like this
    >>> from pandas import DataFrame
    >>> # can be expressed as follows:
    >>> import_object("pandas.DataFrame")
    """

    parts = [part for part in import_path.split(".") if part]
    module, n = None, 0

    while n < len(parts):
        try:
            module = importlib.import_module(".".join(parts[: n + 1]))
            n = n + 1
        except ImportError:
            break

    obj = module or builtins
    for part in parts[n:]:
        obj = getattr(obj, part)

    return obj



from __future__ import annotations

import itertools
import json
import re
import textwrap
import warnings
from typing import Any, Iterable

import pandas as pd
import sqlalchemy as sa
import sqlalchemy.exc
import sqlalchemy.sql.elements

from pydiverse.pipedag import Stage, Table
from pydiverse.pipedag.backend.table.base import BaseTableStore, TableHook
from pydiverse.pipedag.backend.table.util import engine_dispatch
from pydiverse.pipedag.backend.table.util.sql_ddl import (
    AddIndex,
    AddPrimaryKey,
    ChangeColumnNullable,
    ChangeColumnTypes,
    CopyTable,
    CreateDatabase,
    CreateSchema,
    CreateTableAsSelect,
    CreateViewAsSelect,
    DropFunction,
    DropProcedure,
    DropSchema,
    DropTable,
    DropView,
    RenameSchema,
    Schema,
    split_ddl_statement,
)
from pydiverse.pipedag.context import RunContext
from pydiverse.pipedag.context.context import ConfigContext, StageCommitTechnique
from pydiverse.pipedag.errors import CacheError
from pydiverse.pipedag.materialize.container import RawSql
from pydiverse.pipedag.materialize.core import MaterializingTask
from pydiverse.pipedag.materialize.metadata import (
    LazyTableMetadata,
    RawSqlMetadata,
    TaskMetadata,
)
from pydiverse.pipedag.materialize.util import compute_cache_key


class SQLTableStore(BaseTableStore):
    """Table store that materializes tables to a SQL database

    Uses schema swapping for transactions:
    Creates a schema for each stage and a temporary schema for each
    transaction. If all tasks inside a stage succeed, swaps the schemas by
    renaming them.
    """

    METADATA_SCHEMA = "pipedag_metadata"

    @classmethod
    def _init_conf_(cls, config: dict[str, Any]):
        engine_url = config.pop("url")
        return cls(engine_url, **config)

    def __init__(
        self,
        engine_url: str,
        create_database_if_not_exists: bool = False,
        schema_prefix: str = "",
        schema_suffix: str = "",
        disable_pytsql: bool = False,
        pytsql_isolate_top_level_statements: bool = True,
        print_materialize: bool = False,
        print_sql: bool = False,
        no_db_locking: bool = True,
    ):
        """
        Construct table store.

        :param engine_url:
            URL for SQLAlchemy engine
        :param create_database_if_not_exists:
            whether to create database if it does not exist
        :param schema_prefix:
            prefix string for schemas (dot is interpreted as database.schema)
        :param schema_suffix:
            suffix string for schemas (dot is interpreted as database.schema)
        :param disable_pytsql:
            whether to disable the use of pytsql (dialect mssql only)
        :param pytsql_isolate_top_level_statements:
            forward pytsql executes() parameter
        :param print_materialize:
            whether to print select statements before materialization
        :param print_sql:
            whether to print final SQL statements (except for metadata)
        :param no_db_locking:
            speed up database by telling it we will not rely on it's locking mechanisms
        """
        super().__init__()

        self.create_database_if_not_exists = create_database_if_not_exists
        self.schema_prefix = schema_prefix
        self.schema_suffix = schema_suffix
        self.disable_pytsql = disable_pytsql
        self.pytsql_isolate_top_level_statements = pytsql_isolate_top_level_statements
        self.print_materialize = print_materialize
        self.print_sql = print_sql
        self.no_db_locking = no_db_locking
        self.metadata_schema = self.get_schema(self.METADATA_SCHEMA)

        self._init_database(engine_url, create_database_if_not_exists)
        self.engine = self._connect(engine_url, self.schema_prefix, self.schema_suffix)

        # Set up metadata tables and schema
        from sqlalchemy import BigInteger, Boolean, Column, DateTime, String

        self.sql_metadata = sa.MetaData()

        # Stage Table is unique for stage
        # (we currently cannot version changes of this metadata table when using
        # stage_commit_tequnique=read_views)
        self.stage_table = sa.Table(
            "stages",
            self.sql_metadata,
            Column("id", BigInteger, primary_key=True, autoincrement=True),
            Column("stage", String(64)),
            Column("cur_transaction_name", String(256)),
            schema=self.metadata_schema.get(),
        )

        # Store version number for metadata table schema evolution.
        # We disable caching in case of version mismatch.
        self.disable_caching = False
        self.metadata_version = "0.2.0"  # Increase version if metadata table changes
        self.version_table = sa.Table(
            "metadata_version",
            self.sql_metadata,
            Column("version", String(32)),
            schema=self.metadata_schema.get(),
        )

        # Task Table is unique for stage * in_transaction_schema
        self.tasks_table = sa.Table(
            "tasks",
            self.sql_metadata,
            Column("id", BigInteger, primary_key=True, autoincrement=True),
            Column("name", String(128)),
            Column("stage", String(64)),
            Column("version", String(64)),
            Column("timestamp", DateTime),
            Column("run_id", String(32)),
            Column("input_hash", String(32)),
            Column("cache_fn_hash", String(32)),
            Column("output_json", String(2048)),  # 2k might be too small => TBD
            Column("in_transaction_schema", Boolean),
            schema=self.metadata_schema.get(),
        )

        # Lazy Cache Table is unique for stage * in_transaction_schema
        self.lazy_cache_table = sa.Table(
            "lazy_tables",
            self.sql_metadata,
            Column("id", BigInteger, primary_key=True, autoincrement=True),
            Column("name", String(128)),
            Column("stage", String(64)),
            Column("query_hash", String(32)),
            Column("task_hash", String(32)),
            Column("in_transaction_schema", Boolean),
            schema=self.metadata_schema.get(),
        )

        # Sql Cache Table is unique for stage * in_transaction_schema
        self.raw_sql_cache_table = sa.Table(
            "raw_sql_tables",
            self.sql_metadata,
            Column("id", BigInteger, primary_key=True, autoincrement=True),
            Column("prev_tables", String(256)),
            Column("tables", String(256)),
            Column("stage", String(64)),
            Column("query_hash", String(32)),
            Column("task_hash", String(32)),
            Column("in_transaction_schema", Boolean),
            schema=self.metadata_schema.get(),
        )

    @staticmethod
    def _init_database(engine_url: str, create_database_if_not_exists: bool):
        if not create_database_if_not_exists:
            return

        engine = sa.create_engine(engine_url)
        if engine.dialect.name in ["mssql", "ibm_db_sa"]:
            engine.dispose()
            return

        # Attention: This is a really hacky way to create a generic engine for
        #            creating a database before one can open a connection to
        #            self.engine_url which references a database and will fail
        #            if the database does not exist
        url = sa.engine.make_url(engine_url)

        if engine.dialect.name == "postgresql":
            try:
                with engine.connect() as conn:
                    # try whether connection with database in connect string works
                    conn.execute("SELECT 1")
            except sa.exc.DBAPIError:
                postgres_db_url = url.set(database="postgres")
                tmp_engine = sa.create_engine(postgres_db_url)
                try:
                    with tmp_engine.connect() as conn:
                        conn.execute("COMMIT")
                        conn.execute(CreateDatabase(url.database))
                except sa.exc.DBAPIError:
                    # This happens if multiple instances try to create the database
                    # at the same time.
                    with engine.connect() as conn:
                        # Verify database actually exists
                        conn.execute("SELECT 1")
        else:
            raise NotImplementedError(
                "create_database_if_not_exists is only implemented for postgres, yet"
            )
        engine.dispose()

    @staticmethod
    def _connect(engine_url, schema_prefix, schema_suffix):
        engine = sa.create_engine(engine_url)
        if engine.dialect.name == "mssql":
            engine.dispose()
            # this is needed to allow for CREATE DATABASE statements
            # (we don't rely on database transactions anyways)
            engine = sa.create_engine(engine_url, connect_args={"autocommit": True})

        if engine.dialect.name == "mssql":
            if "." in schema_prefix and "." in schema_suffix:
                raise AttributeError(
                    "Config Error: It is not allowed to have a dot in both"
                    " schema_prefix and schema_suffix for SQL Server / mssql database:"
                    f' schema_prefix="{schema_prefix}", schema_suffix="{schema_suffix}"'
                )

            if (schema_prefix + schema_suffix).count(".") != 1:
                raise AttributeError(
                    "Config Error: There must be exactly dot in both schema_prefix and"
                    " schema_suffix together for SQL Server / mssql database:"
                    f' schema_prefix="{schema_prefix}", schema_suffix="{schema_suffix}"'
                )
        return engine

    def get_schema(self, name):
        return Schema(name, self.schema_prefix, self.schema_suffix)

    def _execute(self, query, conn: sa.engine.Connection):
        if self.print_sql:
            if isinstance(query, str):
                query_str = query
            else:
                query_str = str(
                    query.compile(self.engine, compile_kwargs={"literal_binds": True})
                )
            pretty_query_str = self.format_sql_string(query_str)
            self.logger.info(f"Executing sql:\n{pretty_query_str}")

        return conn.execute(query)

    def execute(self, query, *, conn: sa.engine.Connection = None):
        if conn is None:
            with self.engine.connect() as conn:
                return self.execute(query, conn=conn)

        if isinstance(query, sa.schema.DDLElement):
            # Some custom DDL statements contain multiple statements.
            # They are all seperated using a special seperator.
            query_str = str(
                query.compile(self.engine, compile_kwargs={"literal_binds": True})
            )
            for part in split_ddl_statement(query_str):
                self._execute(part, conn)
            return

        return self._execute(query, conn)

    def add_indexes(self, table: Table, schema: Schema):
        if table.primary_key is not None:
            key = table.primary_key
            if isinstance(key, str):
                key = [key]
            self.add_primary_key(table.name, schema, key)
        if table.indexes is not None:
            for index in table.indexes:
                self.add_index(table.name, schema, index)

    @engine_dispatch
    def add_primary_key(
        self,
        table_name: str,
        schema: Schema,
        key_columns: list[str],
        name: str | None = None,
    ):
        self.execute(
            ChangeColumnNullable(table_name, schema, key_columns, nullable=False)
        )
        self.execute(AddPrimaryKey(table_name, schema, key_columns, name))

    @engine_dispatch
    def add_index(
        self,
        table_name: str,
        schema: Schema,
        index_columns: list[str],
        name: str | None = None,
    ):
        self.execute(AddIndex(table_name, schema, index_columns, name))

    @add_primary_key.dialect("mssql")
    def _add_primary_key_mssql(
        self,
        table_name: str,
        schema: Schema,
        key_columns: list[str],
        name: str | None = None,
    ):
        sql_types = self.reflect_sql_types(key_columns, table_name, schema)
        # impose some varchar(max) limit to allow use in primary key / index
        # TODO: consider making cap_varchar_max a config option
        self.execute(
            ChangeColumnTypes(
                table_name,
                schema,
                key_columns,
                sql_types,
                nullable=False,
                cap_varchar_max=1024,
            )
        )
        self.execute(AddPrimaryKey(table_name, schema, key_columns, name))

    @add_index.dialect("mssql")
    def _add_index_mssql(
        self, table_name: str, schema: Schema, index: list[str], name: str | None = None
    ):
        sql_types = self.reflect_sql_types(index, table_name, schema)
        if any(
            [
                isinstance(_type, sa.String) and _type.length is None
                for _type in sql_types
            ]
        ):
            # impose some varchar(max) limit to allow use in primary key / index
            self.execute(
                ChangeColumnTypes(
                    table_name, schema, index, sql_types, cap_varchar_max=1024
                )
            )
        self.execute(AddIndex(table_name, schema, index, name))

    # @add_primary_key.dialect("ibm_db_sa")
    # def _add_primary_key_ibm_db_sa(self, table_name: str, schema: Schema,
    #         key_columns: list[str], name: str | None = None):
    #     # # Failed to make this work: (Database hangs when creating index on
    #     # # altered column)
    #     # sql_types = self.reflect_sql_types(key, table, schema)
    #     # if any([isinstance(_type, sa.String) and (_type.length is None or
    #     #           _type.length > 256) for _type in sql_types]):
    #     #     # impose some varchar length limit to allow use in primary key
    #     #     # / index
    #     #     self.execute(ChangeColumnTypes(table.name, schema, key,
    #     #           sql_types, nullable=False, cap_varchar_max=256))
    #     # else:
    #     self.execute(ChangeColumnNullable(table_name, schema, key_columns,
    #           nullable=False))
    #     self.execute(AddPrimaryKey(table_name, schema, key_columns, name))
    #
    # @add_index.dialect("mssql")
    # def _add_index_ibm_db_sa(self, table_name: str, schema: Schema,
    #         index: list[str], name: str | None = None):
    #     # # Failed to make this work: (Database hangs when creating index on
    #     # # altered column)
    #     # sql_types = self.reflect_sql_types(index, table, schema)
    #     # if any([isinstance(_type, sa.String) and (_type.length is None or
    #     #       _type.length > 256) for _type in sql_types]):
    #     #     # impose some varchar(max) limit to allow use in primary key
    #     #     # / index
    #     #     self.execute(
    #     #         ChangeColumnTypes(table.name, schema, index, sql_types,
    #     #               cap_varchar_max=256))
    #     self.execute(AddIndex(table_name, schema, index, name))

    def copy_indexes(
        self, src_table: str, src_schema: Schema, dest_table: str, dest_schema: Schema
    ):
        inspector = sa.inspect(self.engine)
        pk_constraint = inspector.get_pk_constraint(src_table, schema=src_schema.get())
        if len(pk_constraint["constrained_columns"]) > 0:
            self.add_primary_key(
                dest_table,
                dest_schema,
                pk_constraint["constrained_columns"],
                pk_constraint["name"],
            )
        indexes = inspector.get_indexes(src_table, schema=src_schema.get())
        for index in indexes:
            if len(index["include_columns"]) > 0:
                self.logger.warning(
                    "Perfect index recreation in caching is not net implemented",
                    table=dest_table,
                    schema=dest_schema.get(),
                    include_columns=index["include_columns"],
                )
            if any(
                not isinstance(val, list) or len(val) > 0
                for val in index["dialect_options"].values()
            ):
                self.logger.warning(
                    "Perfect index recreation in caching is not net implemented",
                    table=dest_table,
                    schema=dest_schema.get(),
                    dialect_options=index["dialect_options"],
                )
            self.add_index(
                dest_table, dest_schema, index["column_names"], index["name"]
            )

    def reflect_sql_types(
        self, col_names: Iterable[str], table_name: str, schema: Schema
    ):
        meta = sa.MetaData()
        meta.reflect(bind=self.engine, schema=schema.get())
        tables = {table.name: table for table in meta.tables.values()}
        sql_types = [tables[table_name].c[col].type for col in col_names]
        return sql_types

    @staticmethod
    def format_sql_string(query_str: str) -> str:
        return textwrap.dedent(query_str).strip()

    @engine_dispatch
    def execute_raw_sql(self, raw_sql: RawSql):
        """Executed raw SQL statements in the associated transaction stage"""
        for statement in raw_sql.sql.split(";"):
            self.execute(statement)

    @execute_raw_sql.dialect("mssql")
    def _execute_raw_sql_mssql(self, raw_sql: RawSql):
        if self.disable_pytsql:
            return self._execute_raw_sql_mssql_fallback(raw_sql)
        return self._execute_raw_sql_mssql_pytsql(raw_sql)

    def _execute_raw_sql_mssql_fallback(self, raw_sql: RawSql):
        """Alternative to using pytsql for splitting SQL statement.

        Known Problems:
        - DECLARE statements will not be available across GO
        """
        last_use_statement = None
        for statement in re.split(r"\bGO\b", raw_sql.sql, flags=re.IGNORECASE):
            statement = statement.strip()
            if statement == "":
                # allow GO at end of script
                continue

            if re.match(r"\bUSE\b", statement):
                last_use_statement = statement

            with self.engine.connect() as conn:
                if last_use_statement is not None:
                    self.execute(last_use_statement, conn=conn)
                self.execute(statement, conn=conn)

    def _execute_raw_sql_mssql_pytsql(self, raw_sql: RawSql):
        """Use pytsql for executing T-SQL scripts containing many GO statements."""
        # noinspection PyUnresolvedReferences,PyPackageRequirements
        import pytsql

        sql_string = raw_sql.sql
        if self.print_sql:
            print_query_string = self.format_sql_string(sql_string)
            max_length = 5000
            if len(print_query_string) >= max_length:
                print_query_string = print_query_string[:max_length] + " [...]"
            self.logger.info(f"Executing sql:\n{print_query_string}")

        pytsql.executes(
            sql_string,
            self.engine,
            isolate_top_level_statements=self.pytsql_isolate_top_level_statements,
        )

    def setup(self):
        super().setup()
        self.execute(CreateSchema(self.metadata_schema, if_not_exists=True))
        with self.engine.connect() as conn:
            try:
                version = conn.execute(
                    sa.select(self.version_table.c.version)
                ).scalar_one_or_none()
            except sa.exc.ProgrammingError:
                # table metadata_version does not yet exist
                version = None
            if version is None:
                with conn.begin():
                    self.sql_metadata.create_all(conn)
                    version = conn.execute(
                        sa.select(self.version_table.c.version)
                    ).scalar_one_or_none()
                    assert version is None  # detect race condition
                    conn.execute(
                        self.version_table.insert().values(
                            version=self.metadata_version,
                        )
                    )
            elif version != self.metadata_version:
                # disable caching due to incompatible metadata table schemas
                # (in future versions, consider automatic metadata schema upgrade)
                self.disable_caching = True
                self.logger.warning(
                    "Disabled caching due to metadata version mismatch",
                    version=version,
                    expected_version=self.metadata_version,
                )

    def dispose(self):
        self.engine.dispose()
        super().dispose()

    def init_stage(self, stage: Stage):
        stage_commit_technique = ConfigContext.get().stage_commit_technique
        if stage_commit_technique == StageCommitTechnique.SCHEMA_SWAP:
            return self._init_stage_schema_swap(stage)
        if stage_commit_technique == StageCommitTechnique.READ_VIEWS:
            return self._init_stage_read_views(stage)
        raise ValueError(f"Invalid stage commit technique: {stage_commit_technique}")

    @engine_dispatch
    def _init_stage_schema_swap(self, stage: Stage):
        schema = self.get_schema(stage.name)
        transaction_schema = self.get_schema(stage.transaction_name)

        cs_base = CreateSchema(schema, if_not_exists=True)
        ds_trans = DropSchema(
            transaction_schema, if_exists=True, cascade=True, engine=self.engine
        )
        cs_trans = CreateSchema(transaction_schema, if_not_exists=False)

        with self.engine.connect() as conn:
            self.execute(cs_base, conn=conn)
            self.execute(ds_trans, conn=conn)
            self.execute(cs_trans, conn=conn)

            if not self.disable_caching:
                conn.execute(
                    self.tasks_table.delete()
                    .where(self.tasks_table.c.stage == stage.name)
                    .where(self.tasks_table.c.in_transaction_schema.in_([True]))
                )

    @_init_stage_schema_swap.dialect("mssql")
    def _init_stage_schema_swap_mssql(self, stage: Stage):
        if "." not in self.schema_suffix:
            return self._init_stage_schema_swap.original(self, stage)

        schema = self.get_schema(stage.name)
        transaction_schema = self.get_schema(stage.transaction_name)

        cs_base = CreateSchema(schema, if_not_exists=True)
        cs_trans = CreateSchema(transaction_schema, if_not_exists=True)

        # TODO: detect whether tmp_schema exists and rename it as base or
        #       transaction schema if one is missing
        database, trans_schema_only = transaction_schema.get().split(".")

        # don't drop/create databases, just replace the schema underneath
        # (files will keep name on renaming)
        with self.engine.connect() as conn:
            self.execute(cs_base, conn=conn)
            self.execute(cs_trans, conn=conn)
            self.execute(f"USE [{database}]", conn=conn)
            # clear tables in schema
            self.execute(
                f"""
                EXEC sp_MSforeachtable
                  @command1 = 'DROP TABLE ?'
                , @whereand = 'AND SCHEMA_NAME(schema_id) = ''{trans_schema_only}'' '
                """,
                conn=conn,
            )

            # clear views and stored procedures
            # the 'USE [{database}]' statement is important for these calls
            for view in self.get_view_names(transaction_schema.get()):
                self.execute(
                    DropView(view, transaction_schema, if_exists=True), conn=conn
                )

            modules = self._get_mssql_sql_modules(transaction_schema.get())
            for name, _type in modules.items():
                _type = _type.strip()
                if _type == "P":
                    self.execute(
                        DropProcedure(name, transaction_schema, if_exists=True),
                        conn=conn,
                    )
                elif _type == "FN":
                    self.execute(
                        DropFunction(name, transaction_schema, if_exists=True),
                        conn=conn,
                    )

            # Clear metadata
            if not self.disable_caching:
                conn.execute(
                    self.tasks_table.delete()
                    .where(self.tasks_table.c.stage == stage.name)
                    .where(self.tasks_table.c.in_transaction_schema.in_([True]))
                )

    def _init_stage_read_views(self, stage: Stage):
        try:
            with self.engine.connect() as conn:
                metadata_rows = (
                    conn.execute(
                        self.stage_table.select().where(
                            self.stage_table.c.stage == stage.name
                        )
                    )
                    .mappings()
                    .one()
                )
            cur_transaction_name = metadata_rows["cur_transaction_name"]
        except (sa.exc.MultipleResultsFound, sa.exc.NoResultFound):
            cur_transaction_name = ""

        suffix = "__even" if cur_transaction_name.endswith("__odd") else "__odd"
        new_transaction_name = stage.name + suffix

        # TODO: WE SHOULD NOT MODIFY THE STAGE AFTER CREATION!!!
        stage.set_transaction_name(new_transaction_name)

        # Call schema swap function with updated transaction name
        self._init_stage_schema_swap(stage)

    def commit_stage(self, stage: Stage):
        stage_commit_technique = ConfigContext.get().stage_commit_technique
        if stage_commit_technique == StageCommitTechnique.SCHEMA_SWAP:
            return self._commit_stage_schema_swap(stage)
        if stage_commit_technique == StageCommitTechnique.READ_VIEWS:
            return self._commit_stage_read_views(stage)
        raise ValueError(f"Invalid stage commit technique: {stage_commit_technique}")

    def _commit_stage_schema_swap(self, stage: Stage):
        schema = self.get_schema(stage.name)
        transaction_schema = self.get_schema(stage.transaction_name)
        tmp_schema = self.get_schema(stage.name + "__swap")

        # potentially this disposal must be optional since it does not allow
        # for multithreaded stage execution
        # dispose open connections which may prevent schema swapping
        self.engine.dispose()

        with self.engine.connect() as conn, conn.begin():
            # TODO: for mssql try to find schema does not exist and then move
            #       the forgotten tmp schema there
            self.execute(
                DropSchema(tmp_schema, if_exists=True, cascade=True), conn=conn
            )
            # TODO: in case "." is in self.schema_prefix, we need to implement
            #       schema renaming by creating the new schema and moving
            #       table objects over
            self.execute(RenameSchema(schema, tmp_schema), conn=conn)
            self.execute(RenameSchema(transaction_schema, schema), conn=conn)
            self.execute(RenameSchema(tmp_schema, transaction_schema), conn=conn)

            self._commit_stage_update_metadata(stage, conn=conn)

    def _commit_stage_read_views(self, stage: Stage):
        dest_schema = self.get_schema(stage.name)
        src_schema = self.get_schema(stage.transaction_name)

        with self.engine.connect() as conn:
            # Delete all read views in visible (destination) schema
            views = self.get_view_names(dest_schema.get())
            for view in views:
                self.execute(DropView(view, schema=dest_schema), conn=conn)

            # Create views for all tables in transaction schema
            src_meta = sa.MetaData()
            src_meta.reflect(bind=self.engine, schema=src_schema.get())
            for table in src_meta.tables.values():
                self.execute(
                    CreateViewAsSelect(table.name, dest_schema, sa.select(table)),
                    conn=conn,
                )

            # Update metadata
            stage_metadata_exists = (
                conn.execute(
                    sa.select(1).where(self.stage_table.c.stage == stage.name)
                ).scalar()
                == 1
            )

            if stage_metadata_exists:
                conn.execute(
                    self.stage_table.update()
                    .where(self.stage_table.c.stage == stage.name)
                    .values(cur_transaction_name=stage.transaction_name)
                )
            else:
                conn.execute(
                    self.stage_table.insert().values(
                        stage=stage.name,
                        cur_transaction_name=stage.transaction_name,
                    )
                )

            self._commit_stage_update_metadata(stage, conn=conn)

    def _commit_stage_update_metadata(self, stage: Stage, conn: sa.engine.Connection):
        if not self.disable_caching:
            for table in [
                self.tasks_table,
                self.lazy_cache_table,
                self.raw_sql_cache_table,
            ]:
                conn.execute(
                    table.delete()
                    .where(table.c.stage == stage.name)
                    .where(table.c.in_transaction_schema.in_([False]))
                )
                conn.execute(
                    table.update()
                    .where(table.c.stage == stage.name)
                    .values(in_transaction_schema=False)
                )

    def copy_table_to_transaction(self, table: Table):
        stage = table.stage
        schema_name = self.get_schema(stage.name).get()
        has_table = sa.inspect(self.engine).has_table(table.name, schema=schema_name)
        if not has_table:
            raise CacheError(
                f"Can't copy table '{table.name}' (schema: '{stage.name}')"
                " to transaction because no such table exists."
            )

        try:
            self.execute(
                CopyTable(
                    table.name,
                    self.get_schema(stage.name),
                    table.name,
                    self.get_schema(stage.transaction_name),
                )
            )
        except Exception as _e:
            msg = (
                f"Failed to copy table '{table.name}' (schema: '{stage.name}')"
                " to transaction."
            )
            raise CacheError(msg) from _e
        self.add_indexes(table, self.get_schema(stage.transaction_name))

    def copy_lazy_table_to_transaction(self, metadata: LazyTableMetadata, table: Table):
        stage = table.stage
        schema_name = self.get_schema(metadata.stage).get()
        has_table = sa.inspect(self.engine).has_table(metadata.name, schema=schema_name)
        if not has_table:
            raise CacheError(
                f"Can't copy lazy table '{metadata.name}' (schema:"
                f" '{metadata.stage}') to transaction because no such table"
                " exists."
            )

        try:
            self.execute(
                CopyTable(
                    metadata.name,
                    self.get_schema(metadata.stage),
                    metadata.name,
                    self.get_schema(stage.transaction_name),
                )
            )
        except Exception as _e:
            msg = (
                f"Failed to copy lazy table {metadata.name} (schema:"
                f" '{metadata.stage}') to transaction."
            )
            raise CacheError(msg) from _e
        self.add_indexes(table, self.get_schema(stage.transaction_name))

    @engine_dispatch
    def get_view_names(self, schema: str, *, include_everything=False) -> list[str]:
        """
        List all views that are present in a schema.

        It may also include other database objects like stored procedures, functions,
        etc. which makes the name `get_view_names` too specific. But sqlalchemy
        only allows reading views for all dialects thus the storyline of dialect
        agnostic callers is much nicer to read. In the end we might need everything
        to recover the full cache output which was produced by a RawSQL statement
        (we want to be compatible with legacy sql code as a starting point).
        :param schema: the schema
        :param include_everything: If True, we might include stored procedures,
            functions and other database objects that have a schema associated name.
            Currently, this only makes a difference for dialect=mssql.
        :return: list of view names [and other objects]
        """
        _ = include_everything  # not used in this implementation
        inspector = sa.inspect(self.engine)
        return inspector.get_view_names(schema)

    @get_view_names.dialect("mssql")
    def _get_view_names_mssql(self, schema: str, *, include_everything=False):
        if not include_everything:
            return self.get_view_names.original(
                self, schema, include_everything=include_everything
            )
        return list(self._get_mssql_sql_modules(schema).keys())

    # noinspection SqlDialectInspection
    def _get_mssql_sql_modules(self, schema: str):
        with self.engine.connect() as conn:
            database, schema_only = schema.split(".")
            self.execute(f"USE {database}", conn=conn)
            # include stored procedures in view names because they can also be recreated
            # based on sys.sql_modules.description
            sql = (
                "select obj.name, obj.type from sys.sql_modules as mod "
                "left join sys.objects as obj on mod.object_id=obj.object_id "
                "left join sys.schemas as schem on schem.schema_id=obj.schema_id "
                f"where schem.name='{schema_only}'"
            )
            rows = self.execute(sql, conn=conn).fetchall()
        return {row[0]: row[1] for row in rows}

    def copy_raw_sql_tables_to_transaction(
        self, metadata: RawSqlMetadata, target_stage: Stage
    ):
        src_schema = self.get_schema(metadata.stage)
        dest_schema = self.get_schema(target_stage.transaction_name)

        views = set(self.get_view_names(src_schema.get(), include_everything=True))
        new_tables = set(metadata.tables) - set(metadata.prev_tables)

        tables_to_copy = new_tables - set(views)
        for table_name in tables_to_copy:
            try:
                self.execute(
                    CopyTable(
                        table_name,
                        src_schema,
                        table_name,
                        dest_schema,
                    )
                )
                self.copy_indexes(
                    table_name,
                    src_schema,
                    table_name,
                    dest_schema,
                )
            except Exception as _e:
                msg = (
                    f"Failed to copy table {table_name} (schema:"
                    f" '{src_schema}') to transaction."
                )
                raise CacheError(msg) from _e

        views_to_copy = new_tables & set(views)
        for view_name in views_to_copy:
            self._copy_view_to_transaction(view_name, src_schema, dest_schema)

    @engine_dispatch
    def _copy_view_to_transaction(
        self, view_name: str, src_schema: Schema, dest_schema: Schema
    ):
        raise NotImplementedError("Only implemented for mssql.")

    # noinspection SqlDialectInspection
    @_copy_view_to_transaction.dialect("mssql")
    def _copy_view_to_transaction_mssql(
        self, view_name: str, src_schema: Schema, dest_schema: Schema
    ):
        src_database, src_schema_only = src_schema.get().split(".")
        dest_database, dest_schema_only = dest_schema.get().split(".")

        with self.engine.connect() as conn:
            self.execute(f"USE {src_database}", conn=conn)
            view_sql = self.execute(
                f"""
                SELECT definition
                FROM sys.sql_modules
                WHERE [object_id] = OBJECT_ID('[{src_schema_only}].[{view_name}]');
                """,
                conn=conn,
            ).first()

            if view_sql is None:
                msg = f"No SQL query associated with view {view_name} found."
                raise ValueError(msg)
            view_sql = view_sql[0]

            # Update view SQL so that it references the destination schema
            if src_schema_only != dest_schema_only:
                names_product = itertools.product(
                    [src_schema_only, f"[{src_schema_only}]"],
                    [view_name, f"[{view_name}]"],
                )
                for schema, view in names_product:
                    view_sql.replace(
                        f"{schema}.{view}",
                        f"[{dest_schema_only}].[{view_name}]",
                    )

            self.execute(f"USE {dest_database}", conn=conn)
            self.execute(view_sql, conn=conn)

    def delete_table_from_transaction(self, table: Table):
        self.execute(
            DropTable(
                table.name,
                self.get_schema(table.stage.transaction_name),
                if_exists=True,
            )
        )

    def store_task_metadata(self, metadata: TaskMetadata, stage: Stage):
        if not self.disable_caching:
            with self.engine.connect() as conn:
                conn.execute(
                    self.tasks_table.insert().values(
                        name=metadata.name,
                        stage=metadata.stage,
                        version=metadata.version,
                        timestamp=metadata.timestamp,
                        run_id=metadata.run_id,
                        input_hash=metadata.input_hash,
                        cache_fn_hash=metadata.cache_fn_hash,
                        output_json=metadata.output_json,
                        in_transaction_schema=True,
                    )
                )

    def retrieve_task_metadata(
        self, task: MaterializingTask, input_hash: str, cache_fn_hash: str
    ) -> TaskMetadata:
        if self.disable_caching:
            raise CacheError(
                "Caching is disabled, so we also don't even try to retrieve task"
                f" cache: {task}"
            )

        ignore_fresh_input = RunContext.get().ignore_fresh_input
        try:
            with self.engine.connect() as conn:
                result = (
                    conn.execute(
                        self.tasks_table.select()
                        .where(self.tasks_table.c.name == task.name)
                        .where(self.tasks_table.c.stage == task.stage.name)
                        .where(self.tasks_table.c.version == task.version)
                        .where(self.tasks_table.c.input_hash == input_hash)
                        .where(
                            self.tasks_table.c.cache_fn_hash == cache_fn_hash
                            if not ignore_fresh_input
                            else sa.literal(True)
                        )
                        .where(self.tasks_table.c.in_transaction_schema.in_([False]))
                    )
                    .mappings()
                    .one_or_none()
                )
        except sa.exc.MultipleResultsFound:
            raise CacheError("Multiple results found task metadata") from None

        if result is None:
            raise CacheError(f"Couldn't retrieve task from cache: {task}")

        return TaskMetadata(
            name=result.name,
            stage=result.stage,
            version=result.version,
            timestamp=result.timestamp,
            run_id=result.run_id,
            input_hash=result.input_hash,
            cache_fn_hash=result.cache_fn_hash,
            output_json=result.output_json,
        )

    def store_lazy_table_metadata(self, metadata: LazyTableMetadata):
        if not self.disable_caching:
            with self.engine.connect() as conn:
                conn.execute(
                    self.lazy_cache_table.insert().values(
                        name=metadata.name,
                        stage=metadata.stage,
                        query_hash=metadata.query_hash,
                        task_hash=metadata.task_hash,
                        in_transaction_schema=True,
                    )
                )

    # noinspection DuplicatedCode
    def retrieve_lazy_table_metadata(
        self, query_hash: str, task_hash: str, stage: Stage
    ) -> LazyTableMetadata:
        if self.disable_caching:
            raise CacheError(
                "Caching is disabled, so we also don't even try to retrieve lazy table"
                " cache"
            )

        try:
            with self.engine.connect() as conn:
                result = (
                    conn.execute(
                        self.lazy_cache_table.select()
                        .where(self.lazy_cache_table.c.stage == stage.name)
                        .where(self.lazy_cache_table.c.query_hash == query_hash)
                        .where(self.lazy_cache_table.c.task_hash == task_hash)
                        .where(
                            self.lazy_cache_table.c.in_transaction_schema.in_([False])
                        )
                    )
                    .mappings()
                    .one_or_none()
                )
        except sa.exc.MultipleResultsFound:
            raise CacheError(
                "Multiple results found for lazy table cache key"
            ) from None

        if result is None:
            raise CacheError("No result found for lazy table cache key")

        return LazyTableMetadata(
            name=result.name,
            stage=result.stage,
            query_hash=result.query_hash,
            task_hash=result.task_hash,
        )

    def store_raw_sql_metadata(self, metadata: RawSqlMetadata):
        if not self.disable_caching:
            with self.engine.connect() as conn:
                conn.execute(
                    self.raw_sql_cache_table.insert().values(
                        prev_tables=json.dumps(metadata.prev_tables),
                        tables=json.dumps(metadata.tables),
                        stage=metadata.stage,
                        query_hash=metadata.query_hash,
                        task_hash=metadata.task_hash,
                        in_transaction_schema=True,
                    )
                )

    # noinspection DuplicatedCode
    def retrieve_raw_sql_metadata(
        self, query_hash: str, task_hash: str, stage: Stage
    ) -> RawSqlMetadata:
        if self.disable_caching:
            raise CacheError(
                "Caching is disabled, so we also don't even try to retrieve raw sql"
                " cache"
            )

        try:
            with self.engine.connect() as conn:
                result = (
                    conn.execute(
                        self.raw_sql_cache_table.select()
                        .where(self.raw_sql_cache_table.c.stage == stage.name)
                        .where(self.raw_sql_cache_table.c.query_hash == query_hash)
                        .where(self.raw_sql_cache_table.c.task_hash == task_hash)
                        .where(
                            self.raw_sql_cache_table.c.in_transaction_schema.in_(
                                [False]
                            )
                        )
                    )
                    .mappings()
                    .one_or_none()
                )
        except sa.exc.MultipleResultsFound:
            raise CacheError("Multiple results found for raw sql cache key") from None

        if result is None:
            raise CacheError("No result found for raw sql cache key")

        return RawSqlMetadata(
            prev_tables=json.loads(result.prev_tables),
            tables=json.loads(result.tables),
            stage=result.stage,
            query_hash=result.query_hash,
            task_hash=result.task_hash,
        )

    def list_tables(self, stage: Stage, *, include_everything=False):
        """
        List all tables that were generated in a stage.

        It may also include other objects database objects like views, stored
        procedures, functions, etc. which makes the name `list_tables` too specific.
        But the predominant idea is that tasks produce tables in stages and thus the
        storyline of callers is much nicer to read. In the end we might need everything
        to recover the full cache output which was produced by a RawSQL statement
        (we want to be compatible with legacy sql code as a starting point).

        :param stage: the stage
        :param include_everything: If True, we might include stored procedures,
            functions and other database objects that have a schema associated name.
        :return: list of tables [and other objects]
        """
        inspector = sa.inspect(self.engine)
        schema = self.get_schema(stage.transaction_name).get()

        table_names = inspector.get_table_names(schema)
        view_names = self.get_view_names(schema, include_everything=include_everything)

        return table_names + view_names

    def get_stage_hash(self, stage: Stage) -> str:
        """Compute hash that represents entire stage's output metadata."""
        # We only need to look in tasks_table since the output_json column is updated
        # after evaluating lazy output objects for cache validity. This is the same
        # information we use for producing input_hash of downstream tasks.
        if self.disable_caching:
            raise NotImplementedError(
                "computing stage hash with disabled caching is currently not supported"
            )
        with self.engine.connect() as conn:
            result = conn.execute(
                sa.select([self.tasks_table.c.output_json])
                .where(self.tasks_table.c.stage == stage.name)
                .where(self.tasks_table.c.in_transaction_schema.in_([False]))
                .order_by(self.tasks_table.c.output_json)
            ).all()
            result = [row[0] for row in result]
        return compute_cache_key(*result)


@SQLTableStore.register_table()
class SQLAlchemyTableHook(TableHook[SQLTableStore]):
    @classmethod
    def can_materialize(cls, type_) -> bool:
        return issubclass(
            type_, (sa.sql.elements.TextClause, sa.sql.selectable.Selectable)
        )

    @classmethod
    def can_retrieve(cls, type_) -> bool:
        return type_ == sa.Table

    @classmethod
    def materialize(
        cls,
        store,
        table: Table[sa.sql.elements.TextClause | sa.Text],
        stage_name,
    ):
        obj = table.obj
        if isinstance(table.obj, sa.Table):
            obj = sa.select(table.obj)

        schema = store.get_schema(stage_name)
        store.execute(CreateTableAsSelect(table.name, schema, obj))
        store.add_indexes(table, schema)

    @classmethod
    def retrieve(cls, store, table, stage_name, as_type):
        return sa.Table(
            table.name,
            sa.MetaData(bind=store.engine),
            schema=store.get_schema(stage_name).get(),
            autoload_with=store.engine,
        )

    @classmethod
    def lazy_query_str(cls, store, obj) -> str:
        return str(obj.compile(store.engine, compile_kwargs={"literal_binds": True}))


@SQLTableStore.register_table(pd)
class PandasTableHook(TableHook[SQLTableStore]):
    @classmethod
    def can_materialize(cls, type_) -> bool:
        return issubclass(type_, pd.DataFrame)

    @classmethod
    def can_retrieve(cls, type_) -> bool:
        return type_ == pd.DataFrame

    @classmethod
    def materialize(cls, store, table: Table[pd.DataFrame], stage_name):
        schema = store.get_schema(stage_name)
        if store.print_materialize:
            store.logger.info(
                f"Writing table '{schema.get()}.{table.name}':\n{table.obj}"
            )
        dtype_map = {}
        if store.engine.dialect.name == "ibm_db_sa":
            # Default string target is CLOB which can't be used for indexing.
            # We could use VARCHAR(32000), but if we change this to VARCHAR(256)
            # for indexed columns, DB2 hangs.
            dtype_map = {
                col: sa.VARCHAR(256)
                for col in table.obj.dtypes.loc[lambda x: x == object].index
            }
        table.obj.to_sql(
            table.name,
            store.engine,
            schema=schema.get(),
            index=False,
            dtype=dtype_map,
        )
        store.add_indexes(table, schema)

    @classmethod
    def retrieve(cls, store, table, stage_name, as_type):
        schema = store.get_schema(stage_name).get()
        with store.engine.connect() as conn:
            df = pd.read_sql_table(table.name, conn, schema=schema)
            return df

    @classmethod
    def auto_table(cls, obj: pd.DataFrame):
        if name := obj.attrs.get("name"):
            return Table(obj, name)
        return super().auto_table(obj)


try:
    import pydiverse.transform as pdt
except ImportError as e:
    warnings.warn(str(e), ImportWarning)
    pdt = None


# noinspection PyUnresolvedReferences, PyProtectedMember
@SQLTableStore.register_table(pdt)
class PydiverseTransformTableHook(TableHook[SQLTableStore]):
    @classmethod
    def can_materialize(cls, type_) -> bool:
        return issubclass(type_, pdt.Table)

    @classmethod
    def can_retrieve(cls, type_) -> bool:
        from pydiverse.transform.eager import PandasTableImpl
        from pydiverse.transform.lazy import SQLTableImpl

        return issubclass(type_, (PandasTableImpl, SQLTableImpl))

    @classmethod
    def materialize(cls, store, table: Table[pdt.Table], stage_name):
        from pydiverse.transform.eager import PandasTableImpl
        from pydiverse.transform.lazy import SQLTableImpl

        t = table.obj
        if isinstance(t._impl, PandasTableImpl):
            from pydiverse.transform.core.verbs import collect

            table.obj = t >> collect()
            return PandasTableHook.materialize(store, table, stage_name)
        if isinstance(t._impl, SQLTableImpl):
            table.obj = t._impl.build_select()
            return SQLAlchemyTableHook.materialize(store, table, stage_name)
        raise NotImplementedError

    @classmethod
    def retrieve(cls, store, table, stage_name, as_type):
        from pydiverse.transform.eager import PandasTableImpl
        from pydiverse.transform.lazy import SQLTableImpl

        if issubclass(as_type, PandasTableImpl):
            df = PandasTableHook.retrieve(store, table, stage_name, pd.DataFrame)
            return pdt.Table(PandasTableImpl(table.name, df))
        if issubclass(as_type, SQLTableImpl):
            sa_tbl = SQLAlchemyTableHook.retrieve(store, table, stage_name, sa.Table)
            return pdt.Table(SQLTableImpl(store.engine, sa_tbl))
        raise NotImplementedError

    @classmethod
    def auto_table(cls, obj: pdt.Table):
        return Table(obj, obj._impl.name)

    @classmethod
    def lazy_query_str(cls, store, obj: pdt.Table) -> str:
        from pydiverse.transform.core.verbs import build_query

        query = obj >> build_query()

        if query is not None:
            return str(query)
        return super().lazy_query_str(store, obj)
